
new_theory "more_ltree";
new_parent "List";
new_parent "some_list_thms";


add_theory_to_sml "List";
add_theory_to_sml "some_list_thms";

val LIST_INDUCT_TAC = INDUCT_THEN (theorem "list" "list_INDUCT") ASSUME_TAC;    
val UNDISCH_ONE_TAC =
    let
        fun th_tac (th:thm) (tac:tactic) = (UNDISCH_TAC (concl th)) THEN tac
        fun u_asm  (th:thm) = itlist  th_tac [th] ALL_TAC
    in
        FIRST_ASSUM u_asm
    end;
 
val XRULE_ASSUM_TAC = 
       let 
        fun app (rule:thm->thm) = (fn (t:thm) => (rule t) handle _ =>  t)
       in
       RULE_ASSUM_TAC o app
       end;
  
val MAP_TREE = new_definition ("MAP_TREE",
   (--`MAP_TREE (f:'a->'b) = (@fn. (!v tl. fn (Node v tl) = (Node (f v) (MAP fn tl))))`--));
   
val MAP_TREE_DEF = store_thm("MAP_TREE_DEF",
   (--`!v tl. MAP_TREE (f:'a->'b) (Node v tl) = (Node (f v) (MAP (MAP_TREE f) tl))`--),

    REWRITE_TAC[MAP_TREE] THEN CONV_TAC SELECT_CONV
    THEN MATCH_ACCEPT_TAC 
       (EXISTENCE (BETA_RULE 
            (ISPEC (--`(\(xs:((('b)ltree)list)) (h:'a) (tl:((('a)ltree)list)). (Node (f h) xs))`--)
                   (theorem "ltree" "ltree_Axiom")))));

val Is_ltree_REP_ltree_lemma = store_thm("Is_ltree_REP_ltree_lemma",
 (--`!(f:'a->'b) (t:('a)ltree).
      (Is_ltree ((FST (REP_ltree t)), (MAP f (SND (REP_ltree t)))))`--),
     
  REPEAT STRIP_TAC THEN REWRITE_TAC[definition "ltree" "Is_ltree", theorem "List" "LENGTH_MAP"] 
  THEN STRIP_ASSUME_TAC (SPEC (--`(t:('a)ltree)`--) (prove_abs_fn_onto (definition "ltree" "ltree_ISO_DEF")))
  THEN MP_TAC (SPEC_ALL (CONJUNCT2 (definition "ltree" "ltree_ISO_DEF")))
  THEN ASM_REWRITE_TAC[] THEN DISCH_THEN SUBST1_TAC
  THEN MAP_EVERY POP_ASSUM [MP_TAC, K ALL_TAC]
  THEN ONCE_REWRITE_TAC[SYM (SPEC_ALL (theorem "pair" "PAIR"))]
  THEN REWRITE_TAC[definition "ltree" "Is_ltree"]);

(* Copied from mk_ltree *) 
val size_thm = prove
 (--`!tl. size ((node tl):tree) = SUC(SUM (MAP size tl))`--,
      REWRITE_TAC [definition "ltree" "size"] THEN
      CONV_TAC SELECT_CONV THEN
      MP_TAC (SPEC (--`\n. \(tl:(tree)list). SUC(SUM n)`--) 
                   (INST_TYPE [{redex =  ==`:'b`==, residue = ==`:num`==}]
                              (theorem "tree" "tree_Axiom"))) THEN
      REWRITE_TAC [EXISTS_UNIQUE_DEF] THEN
      CONV_TAC (REDEPTH_CONV BETA_CONV) THEN
      DISCH_THEN (STRIP_THM_THEN CHECK_ASSUME_TAC));
  
(* Copied from mk_ltree *)    
val REP_Node_lemma = store_thm("REP_Node_lemma",
 (--`!tl. REP_ltree (Node (v:'a) tl) = 
         (node (MAP (FST o REP_ltree) tl),
          CONS v (FLAT (MAP (SND o REP_ltree) tl)))`--),
    REWRITE_TAC [(definition "ltree" "Node"), SYM(SPEC_ALL (CONJUNCT2 (definition "ltree" "ltree_ISO_DEF"))), 
                 (definition "ltree" "Is_ltree")] THEN
    LIST_INDUCT_TAC THENL
    [REWRITE_TAC [size_thm, MAP, LENGTH, FLAT, SUM]
    ,
     POP_ASSUM MP_TAC THEN
     REWRITE_TAC [size_thm, MAP, LENGTH, FLAT, SUM, LENGTH_APPEND] THEN
     REWRITE_TAC [SYM (el 4 (CONJUNCTS (theorem "arithmetic" "ADD_CLAUSES")))] THEN
     DISCH_THEN SUBST1_TAC THEN 
     STRIP_TAC THEN STRIP_ASSUME_TAC (SPEC (--`(h:('a)ltree)`--) (prove_abs_fn_onto (definition "ltree" "ltree_ISO_DEF"))) 
     THEN MP_TAC (SPEC (--`r:(tree # ('a)list)`--) (CONJUNCT2 (definition "ltree" "ltree_ISO_DEF"))) THEN
     ASM_REWRITE_TAC [theorem "combin" "o_THM"] THEN
     DISCH_THEN SUBST1_TAC THEN
     MAP_EVERY POP_ASSUM [MP_TAC, K ALL_TAC] THEN
     ONCE_REWRITE_TAC [SYM(SPEC_ALL (theorem "pair" "PAIR"))] THEN
     REWRITE_TAC [(definition "ltree" "Is_ltree")] THEN
     DISCH_THEN SUBST1_TAC THEN REFL_TAC]);


val ltree_ISO_DEF = definition "ltree" "ltree_ISO_DEF";
val EVERY_EL = theorem "list" "EVERY_EL";


val FST_MAP_TREE_TAC =
 ASM_REWRITE_TAC[]
 THEN MP_TAC (ISPEC (--`x:('a)ltree`--) (ISPEC (--`f:'a->'b`--) Is_ltree_REP_ltree_lemma))
 THEN REWRITE_TAC[ltree_ISO_DEF] THEN STRIP_TAC THEN ASM_REWRITE_TAC[theorem "pair" "FST"];

val SND_MAP_TREE_TAC =
 ASM_REWRITE_TAC[]
 THEN MP_TAC (ISPEC (--`x:('a)ltree`--) (ISPEC (--`f:'a->'b`--) Is_ltree_REP_ltree_lemma))
 THEN REWRITE_TAC[ltree_ISO_DEF] THEN STRIP_TAC THEN ASM_REWRITE_TAC[theorem "pair" "SND"];

val MAP_FST_MAP_TREE_TAC =
 IMP_RES_TAC EVERY_EL THEN XRULE_ASSUM_TAC BETA_RULE THEN MATCH_MP_TAC LIST_EQ
 THEN CONJ_TAC THEN REWRITE_TAC[theorem "List" "LENGTH_MAP"]
 THEN REPEAT STRIP_TAC
 THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'a`==, residue = ==`:(('a)ltree)`==}
                             ,{redex = ==`:'b`==, residue = ==`:tree`==}
                             ] EL_MAP)
 THEN ASM_REWRITE_TAC[] THEN BETA_TAC THEN RES_TAC
 THEN ASM_REWRITE_TAC[]
 THEN MP_TAC (ISPEC (--`(EL n l'):('a)ltree`--) (ISPEC (--`f:'a->'b`--) Is_ltree_REP_ltree_lemma))
 THEN REWRITE_TAC[ltree_ISO_DEF] THEN STRIP_TAC THEN ASM_REWRITE_TAC[];
 
val MAP_SND_MAP_TREE_TAC = 
 IMP_RES_TAC EVERY_EL THEN XRULE_ASSUM_TAC BETA_RULE THEN MATCH_MP_TAC LIST_EQ
 THEN CONJ_TAC THEN REWRITE_TAC[theorem "List" "LENGTH_MAP"]
 THEN REPEAT STRIP_TAC
 THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'a`==, residue = ==`:(('a)ltree)`==}
                             ,{redex = ==`:'b`==, residue = ==`:(('b)list)`==}
                             ] EL_MAP)
 THEN ASM_REWRITE_TAC[] THEN BETA_TAC THEN RES_TAC
 THEN ASM_REWRITE_TAC[]
 THEN MP_TAC (ISPEC (--`(EL n l'):('a)ltree`--) (ISPEC (--`f:'a->'b`--) Is_ltree_REP_ltree_lemma))
 THEN REWRITE_TAC[ltree_ISO_DEF] THEN STRIP_TAC THEN ASM_REWRITE_TAC[];

val EVERY_DEF = definition "list" "EVERY_DEF";
val Node = definition "ltree" "Node";

(* Important theorem for the proof of the para *)

val MAP_TREE_ABS = store_thm("MAP_TREE_ABS",   
  (--`!(t:('a)ltree) (f:'a->'b). 
         MAP_TREE f t = (ABS_ltree (FST (REP_ltree t),MAP f (SND (REP_ltree t))))`--),
   
     REPEAT GEN_TAC THEN 
     MATCH_MP_TAC
      (BETA_RULE (SPEC (--`(\(t:('a)ltree). (MAP_TREE (f:'a->'b) t)
                                  = 
                                  (ABS_ltree (FST (REP_ltree t),MAP f (SND (REP_ltree t)))))`--) 
                  (theorem "ltree" "ltree_Induct")))
     THEN GEN_TAC THEN STRIP_ASSUME_TAC (ISPEC (--`t:(('a)ltree)list`--) list_CASES)
     THEN ASM_REWRITE_TAC[EVERY_DEF]
     THENL
      [REWRITE_TAC[MAP_TREE_DEF,REP_Node_lemma,MAP,FLAT,Node] 
      ,
       BETA_TAC THEN REPEAT STRIP_TAC THEN REWRITE_TAC[MAP_TREE_DEF,REP_Node_lemma,MAP,FLAT,Node] 
       THEN REWRITE_TAC[FLAT,MAP_APPEND,theorem "List" "MAP_FLAT",theorem "List" "MAP_MAP_o"]
       THEN REWRITE_TAC[definition "combin" "o_DEF"] THEN BETA_TAC
       
       
       THEN SUBGOAL_THEN (--`(FST (REP_ltree (MAP_TREE (f:'a->'b) x))) = (FST (REP_ltree (x:('a)ltree)))`--)
            ASSUME_TAC
       THENL
       [FST_MAP_TREE_TAC
       ,ALL_TAC]
       THEN SUBGOAL_THEN (--`(MAP (\x. FST (REP_ltree (MAP_TREE (f:'a->'b) x))) (l':(('a)ltree)list)) 
                           = (MAP (\x. FST (REP_ltree x)) l')`--) ASSUME_TAC
       THENL
        [MAP_FST_MAP_TREE_TAC
        ,ALL_TAC]
       THEN SUBGOAL_THEN (--`(SND (REP_ltree (MAP_TREE (f:'a->'b) (x:('a)ltree)))) 
                             = (MAP f (SND (REP_ltree x)))`--) ASSUME_TAC
       THENL
        [SND_MAP_TREE_TAC
        ,ALL_TAC]
       THEN SUBGOAL_THEN (--`(MAP (\x. SND (REP_ltree (MAP_TREE (f:'a->'b) x))) (l':(('a)ltree)list))
                           = (MAP (\x. MAP f (SND (REP_ltree x))) l')`--) ASSUME_TAC
       THENL
        [MAP_SND_MAP_TREE_TAC
        ,ALL_TAC]
       THEN ASM_REWRITE_TAC[]
     ]);



(*-------------------------------------------------*)	 
(* Some operations on ltree							       *)
(*-------------------------------------------------*)


val Tree_Root_DEF = new_definition("Tree_Root_DEF",
   (--`Tree_Root = @fn. (!(v:'a) (tl:(('a)ltree)list). (fn (Node v tl) = v))`--));   
         

val Tree_Root = store_thm("Tree_Root",
  (--`!(v:'a) (tl:(('a)ltree)list). (Tree_Root (Node v tl) = v)`--),
  
  REWRITE_TAC[Tree_Root_DEF] THEN CONV_TAC SELECT_CONV
  THEN MATCH_ACCEPT_TAC 
       (EXISTENCE (BETA_RULE 
            (ISPEC (--`(\(xs:('a)list) (h:'a) (tl:((('a)ltree)list)). h)`--)
                   (theorem "ltree" "ltree_Axiom")))) );

val Tree_Subtrees_DEF = new_definition("Tree_Subtrees_DEF",
   (--`Tree_Subtrees = @fn. (!(v:'a) (tl:(('a)ltree)list). (fn (Node v tl) = tl))`--));   
 

val Tree_Subtrees = store_thm("Tree_Subtrees",
  (--`!(v:'a) (tl:(('a)ltree)list). (Tree_Subtrees (Node v tl) = tl)`--),
  
  REWRITE_TAC[Tree_Subtrees_DEF] THEN CONV_TAC SELECT_CONV
  THEN MATCH_ACCEPT_TAC 
       (EXISTENCE (BETA_RULE 
            (ISPEC (--`(\(xs:((('a)ltree)list)list) (h:'a) (tl:((('a)ltree)list)). tl)`--)
                   (theorem "ltree" "ltree_Axiom")))) );
	 

val MAP_TREE_EQ_lemma = prove(
  (--`!(t:('a)ltree) (f:'a->'b) (g:'a->'b).
      (\t. (((f (Tree_Root t)) = (g (Tree_Root t)))
      /\
       (!(t':('a)ltree). (IS_EL t' (Tree_Subtrees t)) ==> ((MAP_TREE f t') = (MAP_TREE g t'))))
       ==> ((MAP_TREE f t) = (MAP_TREE g t))) t`--),
      
REPEAT GEN_TAC THEN MATCH_MP_TAC (theorem "ltree" "ltree_Induct")
THEN LIST_INDUCT_TAC
THEN BETA_TAC THEN REWRITE_TAC[MAP_TREE_DEF,MAP,Tree_Root,Tree_Subtrees] 
THEN REPEAT STRIP_TAC 
THEN ASM_REWRITE_TAC[theorem "ltree" "Node_11",theorem "list" "CONS_11"]
THEN CONJ_TAC
THENL
[FIRST_ASSUM MATCH_MP_TAC THEN REWRITE_TAC[theorem "List" "IS_EL"]
,
 XRULE_ASSUM_TAC (BETA_RULE o (REWRITE_RULE [EVERY_IS_EL]))
 THEN MATCH_MP_TAC LIST_EQ THEN REWRITE_TAC[theorem "List" "LENGTH_MAP"]
 THEN REPEAT STRIP_TAC
 THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'b`==, residue = ==`:('b)ltree`==}] 
                       (theorem "List" "EL_MAP"))
 THEN ASM_REWRITE_TAC[]  
 THEN SUBGOAL_THEN (--`IS_EL (EL n t) (CONS (h:('a)ltree) t)`--) ASSUME_TAC
 THENL
  [REWRITE_TAC[theorem "List" "IS_EL"] THEN DISJ2_TAC
  THEN MATCH_MP_TAC (theorem "List" "EL_IS_EL") THEN ASM_REWRITE_TAC[]  
  ,
   RES_TAC
  ]
]);


val MAP_TREE_EQ = store_thm("MAP_TREE_EQ",
  (--`!(t:('a)ltree) (f:'a->'b) (g:'a->'b).
      (((f (Tree_Root t)) = (g (Tree_Root t)))
      /\
       (!(t':('a)ltree). (IS_EL t' (Tree_Subtrees t)) ==> ((MAP_TREE f t') = (MAP_TREE g t'))))
       ==> ((MAP_TREE f t) = (MAP_TREE g t))`--),
     MATCH_ACCEPT_TAC (BETA_RULE MAP_TREE_EQ_lemma));





val MAP_TREE_ID_lemma = prove(
  (--`!(t:('a)ltree). (\t. (MAP_TREE (\x. x) t) = t) t`--),
  
MATCH_MP_TAC (theorem "ltree" "ltree_Induct")
THEN BETA_TAC 
THEN REWRITE_TAC[MAP_TREE_DEF,MAP,Tree_Root,Tree_Subtrees,EVERY_IS_EL] 
THEN BETA_TAC THEN REPEAT STRIP_TAC 
THEN ASM_REWRITE_TAC[theorem "ltree" "Node_11",theorem "list" "CONS_11"]
THEN MATCH_MP_TAC LIST_EQ THEN REWRITE_TAC[LENGTH_MAP]
THEN REPEAT STRIP_TAC
THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'a`==, residue = ==`:('a)ltree`==}
                             ,{redex = ==`:'b`==, residue = ==`:('a)ltree`==}] EL_MAP)
THEN ASM_REWRITE_TAC[] THEN FIRST_ASSUM MATCH_MP_TAC
THEN MATCH_MP_TAC EL_IS_EL THEN ASM_REWRITE_TAC[] );

  

val MAP_TREE_ID = store_thm("MAP_TREE_ID",
  (--`!(t:('a)ltree). (MAP_TREE (\x. x) t) = t`--),
  MATCH_ACCEPT_TAC (BETA_RULE MAP_TREE_ID_lemma));
  


val MAP_TREE_o_lemma = prove(
   (--`!(g:'a->'b) (f:'b->'c) (t:('a)ltree).
     (\t. (MAP_TREE f (MAP_TREE g t)) = (MAP_TREE (f o g) t)) t`--),
GEN_TAC THEN GEN_TAC THEN MATCH_MP_TAC (theorem "ltree" "ltree_Induct")
THEN BETA_TAC 
THEN REWRITE_TAC[MAP_TREE_DEF,MAP,Tree_Root,Tree_Subtrees,EVERY_IS_EL] 
THEN BETA_TAC THEN REPEAT STRIP_TAC 
THEN ASM_REWRITE_TAC[theorem "ltree" "Node_11"]
THEN CONJ_TAC
THENL
 [REWRITE_TAC[definition "combin" "o_DEF"] THEN BETA_TAC THEN REFL_TAC
 ,ALL_TAC]
THEN MATCH_MP_TAC LIST_EQ THEN REWRITE_TAC[LENGTH_MAP]
THEN REPEAT STRIP_TAC
THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'a`==, residue = ==`:('a)ltree`==}
                             ,{redex = ==`:'b`==, residue = ==`:('c)ltree`==}] EL_MAP)
THEN ASM_REWRITE_TAC[] 
THEN POP_ASSUM (fn thm => ALL_TAC)
THEN SUBGOAL_THEN (--`n < (LENGTH (MAP (MAP_TREE (g:'a->'b)) t))`--) ASSUME_TAC
THENL[ASM_REWRITE_TAC[LENGTH_MAP],ALL_TAC]
THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'a`==, residue = ==`:('b)ltree`==}
                             ,{redex = ==`:'b`==, residue = ==`:('c)ltree`==}] EL_MAP)
THEN ASM_REWRITE_TAC[] 
THEN POP_ASSUM (fn thm => ALL_TAC)
THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'a`==, residue = ==`:('a)ltree`==}
                             ,{redex = ==`:'b`==, residue = ==`:('b)ltree`==}] EL_MAP)
THEN ASM_REWRITE_TAC[] 
THEN FIRST_ASSUM MATCH_MP_TAC
THEN MATCH_MP_TAC EL_IS_EL THEN ASM_REWRITE_TAC[] );
    
     

val MAP_TREE_o = store_thm("MAP_TREE_o",
   (--`!(g:'a->'b) (f:'b->'c) (t:('a)ltree).
     (MAP_TREE f (MAP_TREE g t)) = (MAP_TREE (f o g) t)`--),
 MATCH_ACCEPT_TAC (BETA_RULE MAP_TREE_o_lemma));
 



val EVERY_TREE = new_definition("EVERY_TREE",
  (--`EVERY_TREE (P:'a->bool) (t:('a)ltree) =
           EVERY P (SND (REP_ltree t))`--));
	   

   
val EVERY_TREE_DEF = store_thm("EVERY_TREE_DEF",
(--`!(P:'a->bool) (h:'a) (t:(('a)ltree)list).
         EVERY_TREE P (Node h t) = (P h) /\ (EVERY (EVERY_TREE P) t)`--),
  REPEAT GEN_TAC THEN REWRITE_TAC[EVERY_TREE, REP_Node_lemma,EVERY_DEF]
  THEN REWRITE_TAC[EVERY_IS_EL,EVERY_TREE,IS_EL_FLAT] THEN EQ_TAC
  THENL
   [REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[] THEN FIRST_ASSUM MATCH_MP_TAC
    THEN EXISTS_TAC (--`(SND (REP_ltree (x:('a)ltree)))`--)
    THEN ASM_REWRITE_TAC[] 
    THEN SUBGOAL_THEN (--`(SND (REP_ltree (x:('a)ltree))) = (SND o REP_ltree) x`--) ASSUME_TAC
    THENL[REWRITE_TAC[definition "combin" "o_DEF"] THEN BETA_TAC THEN REFL_TAC,ALL_TAC]
    THEN ASM_REWRITE_TAC[] THEN MATCH_MP_TAC IS_EL_MAP THEN ASM_REWRITE_TAC[]
   ,
    
    REWRITE_TAC[GSYM IS_EL_MAP_EXISTS] THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[]
    THEN RES_TAC THEN FIRST_ASSUM MATCH_MP_TAC THEN UNDISCH_ONE_TAC
    THEN UNDISCH_ONE_TAC THEN XRULE_ASSUM_TAC SYM 
    THEN ASM_REWRITE_TAC[definition "combin" "o_DEF"] THEN BETA_TAC THEN REPEAT STRIP_TAC
    THEN ASM_REWRITE_TAC[]
   ]);
   

val EVERY_TREE_THM = store_thm("EVERY_TREE_THM",
(--`!(P:'a->bool) (h:'a) (t:('a)ltree).
         EVERY_TREE P t = (P (Tree_Root t)) /\ (EVERY (EVERY_TREE P) (Tree_Subtrees t))`--),
REPEAT GEN_TAC
THEN STRIP_ASSUME_TAC  (ISPEC (--`t:('a)ltree`--) (theorem "ltree" "Node_onto"))
THEN ASM_REWRITE_TAC[Tree_Root, Tree_Subtrees,EVERY_TREE_DEF]);

	 

val EVERY_TREE_IMP_MAP_TREE_EQ_lemma = prove(
 
(--`!(t:('a)ltree) (f:'a->'b) (g:'a->'b).
      (\t. (EVERY_TREE (\x. f x = g x) t)
       ==> ((MAP_TREE f t) = (MAP_TREE g t))) t`--),
       
     
REPEAT GEN_TAC THEN MATCH_MP_TAC (theorem "ltree" "ltree_Induct")
THEN BETA_TAC THEN REWRITE_TAC[MAP_TREE_DEF,MAP,Tree_Root,Tree_Subtrees] 
THEN REPEAT STRIP_TAC 
THEN ASM_REWRITE_TAC[theorem "ltree" "Node_11",theorem "list" "CONS_11"]
THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[EVERY_TREE_DEF]
THEN BETA_TAC
THEN REWRITE_TAC[EVERY_IS_EL]
THEN XRULE_ASSUM_TAC (BETA_RULE o (REWRITE_RULE [EVERY_IS_EL]))
THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[]
THEN MATCH_MP_TAC LIST_EQ THEN REWRITE_TAC[theorem "List" "LENGTH_MAP"]
 THEN REPEAT STRIP_TAC
 THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'b`==, residue = ==`:('b)ltree`==}] 
                       (theorem "List" "EL_MAP"))
 THEN ASM_REWRITE_TAC[]  
 THEN SUBGOAL_THEN (--`IS_EL (EL n t) (t:(('a)ltree)list)`--) ASSUME_TAC
 THENL
  [MATCH_MP_TAC (theorem "List" "EL_IS_EL") THEN ASM_REWRITE_TAC[]  
  ,
   RES_TAC THEN FIRST_ASSUM MATCH_MP_TAC
   THEN XRULE_ASSUM_TAC (BETA_RULE o (REWRITE_RULE [EVERY_TREE_THM]))
   THEN ASM_REWRITE_TAC[]
  ]);
  
  
val EVERY_TREE_IMP_MAP_TREE_EQ = store_thm("EVERY_TREE_IMP_MAP_TREE_EQ",
    (--`!(t:('a)ltree) (f:'a->'b) (g:'a->'b).
       (EVERY_TREE (\x. f x = g x) t)
       ==> ((MAP_TREE f t) = (MAP_TREE g t))`--),
       
MATCH_ACCEPT_TAC (BETA_RULE  EVERY_TREE_IMP_MAP_TREE_EQ_lemma));



val EVERY_TREE_MAP_TREE_o_lemma = prove(
   (--`!(f:'a->'b) (P:'b->bool) (t:('a)ltree).
     (\t. (EVERY_TREE P (MAP_TREE f t)) = (EVERY_TREE (P o f) t)) t`--),

GEN_TAC THEN GEN_TAC THEN MATCH_MP_TAC (theorem "ltree" "ltree_Induct")
THEN BETA_TAC THEN REWRITE_TAC[MAP_TREE_DEF,MAP,Tree_Root,Tree_Subtrees, EVERY_IS_EL,EVERY_TREE_DEF]
THEN REWRITE_TAC[GSYM  IS_EL_MAP_EXISTS]
THEN BETA_TAC THEN REPEAT STRIP_TAC THEN EQ_TAC THEN REPEAT STRIP_TAC
THENL
 [REWRITE_TAC[definition "combin" "o_DEF"] THEN BETA_TAC THEN ASM_REWRITE_TAC[]
 ,
  XRULE_ASSUM_TAC (ISPEC (--`MAP_TREE (f:'a->'b) (x:('a)ltree)`--)) THEN RES_TAC
  THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[] THEN POP_ASSUM (fn thm => ALL_TAC)
  THEN POP_ASSUM (fn thm => ALL_TAC) THEN FIRST_ASSUM MATCH_MP_TAC THEN REFL_TAC
 ,
  XRULE_ASSUM_TAC (BETA_RULE o (REWRITE_RULE[definition "combin" "o_DEF"]))
  THEN ASM_REWRITE_TAC[]
 ,
  XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[] THEN RES_TAC
 ]);
 

val EVERY_TREE_MAP_TREE_o = store_thm("EVERY_TREE_MAP_TREE_o",
   (--`!(f:'a->'b) (P:'b->bool) (t:('a)ltree).
     (EVERY_TREE P (MAP_TREE f t)) = (EVERY_TREE (P o f) t)`--),
   MATCH_ACCEPT_TAC (BETA_RULE EVERY_TREE_MAP_TREE_o_lemma));
   
     


(*------------------------------------------------------*)
(*  ZIP_TREE                                            *)
(*------------------------------------------------------*)


    
val CURRY_ZIP_TREE = new_definition ("CURRY_ZIP_TREE",
   (--`CURRY_ZIP_TREE  = @fn. (!(v:'a) tl. fn (Node v tl) 
                                      = (\(t:('b)ltree). 
                                         Node (v, (Tree_Root t)) 
                                              (MAP (\p. (FST p) (SND p)) (ZIP ((MAP fn tl), (Tree_Subtrees t))))))`--));
                                      
val CURRY_ZIP_TREE_THM = prove (
   (--`!(v:'a) tl. CURRY_ZIP_TREE (Node v tl) = (\(t:('b)ltree). 
                                             Node (v, (Tree_Root t)) 
                                                  (MAP (\p. (FST p) (SND p)) 
                                                     (ZIP ((MAP CURRY_ZIP_TREE tl), (Tree_Subtrees t)))))`--),
   REWRITE_TAC[CURRY_ZIP_TREE] THEN CONV_TAC SELECT_CONV
   THEN MATCH_ACCEPT_TAC 
       (EXISTENCE (BETA_RULE 
            (ISPEC (--`(\(xs:(('b)ltree -> ('a # 'b)ltree)list) (h:'a) (tl:((('a)ltree)list)). 
                        (\(t:('b)ltree). Node (h, (Tree_Root t)) (MAP (\p. (FST p) (SND p)) (ZIP (xs,(Tree_Subtrees t))))))`--)
                   (theorem "ltree" "ltree_Axiom")))) );  

(* ZIP_TREE has the desired property *)

val ZIP_TREE = new_definition ("ZIP_TREE",
  (--`(ZIP_TREE:(('a)ltree # ('b)ltree) -> ('a # 'b)ltree)
         = (UNCURRY CURRY_ZIP_TREE)`--));



(* ZIP_TREE has the desired property *)

val ZIP_TREE_DEF = store_thm("ZIP_TREE_DEF",
  (--`!(v1:'a) (v2:'b) (tl1:(('a)ltree)list) (tl2:(('b)ltree)list).
     ((LENGTH tl1) = (LENGTH tl2)) ==> 
     ((ZIP_TREE ((Node v1 tl1), (Node v2 tl2))) = (Node (v1,v2) (MAP (ZIP_TREE) (ZIP (tl1,tl2)))))`--),
   REWRITE_TAC[definition "pair" "UNCURRY_DEF", ZIP_TREE,CURRY_ZIP_TREE_THM]
   THEN BETA_TAC THEN REWRITE_TAC[Tree_Root, Tree_Subtrees]
   THEN REWRITE_TAC[theorem "ltree" "Node_11"] 
   THEN LIST_INDUCT_TAC 
   THENL
    [LIST_INDUCT_TAC 
     THENL
      [REWRITE_TAC[definition "list" "MAP",definition "List" "ZIP"]		      
      ,
       REWRITE_TAC[theorem "List" "LENGTH",theorem "arithmetic" "SUC_NOT"]
      ]
    , STRIP_TAC THEN LIST_INDUCT_TAC
      THENL
       [REWRITE_TAC[theorem "List" "LENGTH",GSYM (theorem "arithmetic" "SUC_NOT")]
       ,
        REWRITE_TAC[definition "list" "MAP",definition "List" "ZIP"]
        THEN REPEAT STRIP_TAC 
	THEN REWRITE_TAC [theorem "list" "CONS_11", CURRY_ZIP_TREE_THM,definition "pair" "UNCURRY_DEF"]
        THEN BETA_TAC THEN CONJ_TAC
        THENL
         [BETA_TAC THEN REWRITE_TAC[theorem "pair" "FST", theorem "pair" "SND"]
         ,
          FIRST_ASSUM MATCH_MP_TAC THEN UNDISCH_ONE_TAC THEN REWRITE_TAC[theorem "List" "LENGTH"]
          THEN REWRITE_TAC[theorem "prim_rec" "INV_SUC_EQ"]
         ]
       ]
      ]);
      
val ZIP_TREE_EQ = store_thm("ZIP_TREE_EQ",
  (--`!(t1:('a)ltree) (t2:('a)ltree) (b1:('b)ltree) (b2:('b)ltree).
   ((t1 = t2) /\ (b1 = b2)) ==> ((ZIP_TREE (t1, b1)) = (ZIP_TREE (t2, b2)))`--),
   
  REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[]);
 


val Node_11 = theorem "ltree" "Node_11";

val ZIP_TREE_MAP_TREE_EQ_MAP_TREE_splid = store_thm("ZIP_TREE_MAP_TREE_EQ_MAP_TREE_splid",
 
   (--`!(t:('a)ltree) (f:'a->'b). (ZIP_TREE ((MAP_TREE f t), t)) 
                                  = 
                                  (MAP_TREE ((splid f):'a -> ('b # 'a)) t)`--),
 
     REPEAT GEN_TAC THEN 
     MATCH_MP_TAC
      (BETA_RULE (SPEC (--`(\(t:('a)ltree). (ZIP_TREE ((MAP_TREE f t), t)) 
                                  = 
                                  (MAP_TREE ((splid f):'a -> ('b # 'a)) t))`--) (theorem "ltree" "ltree_Induct")))
     THEN GEN_TAC THEN STRIP_ASSUME_TAC (ISPEC (--`t:(('a)ltree)list`--) list_CASES)
     THEN ASM_REWRITE_TAC[definition "list" "EVERY_DEF"] 
     THENL
      [REWRITE_TAC[MAP_TREE_DEF, ZIP_TREE_DEF,splid,MAP]
       THEN SUBGOAL_THEN (--`(LENGTH ([]:(('b)ltree)list)) = (LENGTH ([]:(('a)ltree)list))`--) ASSUME_TAC
       THENL[REWRITE_TAC[LENGTH], ALL_TAC]
       THEN IMP_RES_TAC ZIP_TREE_DEF 
       THEN ASM_REWRITE_TAC[theorem "ltree" "Node_11",MAP,definition "List" "ZIP"]
      ,
       BETA_TAC THEN REPEAT STRIP_TAC
       THEN REWRITE_TAC[MAP_TREE_DEF, ZIP_TREE_DEF,splid,MAP]
       THEN SUBGOAL_THEN (--`(LENGTH ((CONS (MAP_TREE (f:'a->'b) x) (MAP (MAP_TREE f) l')))) = (LENGTH (CONS x l'))`--) ASSUME_TAC
       THENL[REWRITE_TAC[LENGTH_MAP,LENGTH],ALL_TAC]
       THEN IMP_RES_TAC ZIP_TREE_DEF THEN ASM_REWRITE_TAC[Node_11,MAP,definition "List" "ZIP"]
       THEN ASM_REWRITE_TAC[CONS_11] THEN IMP_RES_TAC EVERY_EL THEN XRULE_ASSUM_TAC BETA_RULE 
       THEN MATCH_MP_TAC LIST_EQ
       THEN CONJ_TAC
       THENL
        [REWRITE_TAC[LENGTH_MAP]
        THEN SUBGOAL_THEN (--`LENGTH (MAP (MAP_TREE (f:'a->'b)) (l':(('a)ltree)list)) = LENGTH l'`--) ASSUME_TAC
        THENL[REWRITE_TAC[LENGTH_MAP],ALL_TAC]
        THEN IMP_RES_TAC (theorem "List" "LENGTH_ZIP")
        ,
         REWRITE_TAC[LENGTH_MAP] THEN REPEAT STRIP_TAC
         THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'a`==, residue = ==`:(('c)ltree # ('a)ltree)`==}
                                     ,{redex = ==`:'b`==, residue = ==`:('b # 'a)ltree`==}
                                     ,{redex = ==`:'c`==, residue = ==`:'b`==}] (theorem "List" "EL_MAP"))
         THEN ASM_REWRITE_TAC[]
         THEN SUBGOAL_THEN (--`n < (LENGTH (l':(('a)ltree)list))`--) ASSUME_TAC
         THENL
          [UNDISCH_TAC (--`n < LENGTH (ZIP (MAP (MAP_TREE (f:'a->'b)) (l':(('a)ltree)list),l'))`--)
           THEN SUBGOAL_THEN (--`(LENGTH (MAP (MAP_TREE (f:'a->'b)) (l':(('a)ltree)list))) = (LENGTH l')`--) ASSUME_TAC
           THENL[REWRITE_TAC[LENGTH_MAP],ALL_TAC] THEN IMP_RES_TAC (theorem "List" "LENGTH_ZIP")
           THEN ASM_REWRITE_TAC[]
          ,ALL_TAC]
         THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'a`==, residue = ==`:(('a)ltree)`==}
                                     ,{redex = ==`:'b`==, residue = ==`:('b # 'a)ltree`==}
                                     ] (theorem "List" "EL_MAP"))
         THEN ASM_REWRITE_TAC[] THEN RES_TAC THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[]
         THEN SUBGOAL_THEN (--`EL n (ZIP (MAP (MAP_TREE (f:'a->'b)) (l':(('a)ltree)list),l')) =
                                (EL n (MAP (MAP_TREE f) l'),EL n l')`--) ASSUME_TAC
         THENL
          [MATCH_MP_TAC EL_ZIP THEN ASM_REWRITE_TAC[LENGTH_MAP]
          ,ALL_TAC]
         THEN ASM_REWRITE_TAC[]
         THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'a`==, residue = ==`:(('a)ltree)`==}
                                     ,{redex = ==`:'b`==, residue = ==`:('b)ltree`==}
                                     ] (theorem "List" "EL_MAP"))
         THEN ASM_REWRITE_TAC[]
        ]
      ]);
      

export_theory();

         
