
new_theory "some_list_thms";
new_parent "List";

val LIST_INDUCT_TAC = INDUCT_THEN (theorem "list" "list_INDUCT") ASSUME_TAC;  

 val UNDISCH_ONE_TAC =
    let
        fun th_tac (th:thm) (tac:tactic) = (UNDISCH_TAC (concl th)) THEN tac
        fun u_asm  (th:thm) = itlist  th_tac [th] ALL_TAC
    in
        FIRST_ASSUM u_asm
    end;
 
val XRULE_ASSUM_TAC = 
       let 
        fun app (rule:thm->thm) = (fn (t:thm) => (rule t) handle _ =>  t)
       in
       RULE_ASSUM_TAC o app
       end;

load_library {lib = arith_lib, theory = "-"};
use (my_hol_dir ^ "library/TV_UNITY/src/aux_decls.sml");
add_theory_to_sml "List";

   
(*-------------------------------------------------*)
(* Some theorems about lists                       *)
(*-------------------------------------------------*)   
    
val MAP = definition "list" "MAP";


val IS_EL_MAP   = store_thm("IS_EL_MAP",
  (--`!(xs:('a)list) (x:'a) (f:'a->'b). (IS_EL x xs) ==> (IS_EL (f x) (MAP f xs))`--),
  LIST_INDUCT_TAC
  THEN REWRITE_TAC[theorem "List" "IS_EL",MAP]
  THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[]
  THEN RES_TAC THEN ASM_REWRITE_TAC[]);


val IS_EL_MAP_EXISTS = store_thm("IS_EL_MAP_EXISTS",
  (--`!(xs:('a)list) (f:'a->'b) (y:'b).
          (?(x:'a). (IS_EL x xs) /\ ((f x) = y)) = (IS_EL y (MAP f xs))`--),

LIST_INDUCT_TAC THEN REWRITE_TAC[theorem "List" "IS_EL"]
THEN REWRITE_TAC [theorem "List" "MAP",theorem "List" "IS_EL"]
THEN REPEAT STRIP_TAC THEN EQ_TAC THEN STRIP_TAC
THENL
 [DISJ1_TAC THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[]
 ,
  DISJ2_TAC THEN UNDISCH_ONE_TAC THEN XRULE_ASSUM_TAC GSYM THEN STRIP_TAC
  THEN ASM_REWRITE_TAC[] THEN EXISTS_TAC (--`x:'a`--) THEN ASM_REWRITE_TAC[]
 ,
  EXISTS_TAC (--`h:'a`--) THEN ASM_REWRITE_TAC[]
 ,
  UNDISCH_ONE_TAC THEN XRULE_ASSUM_TAC GSYM THEN ASM_REWRITE_TAC[]
  THEN STRIP_TAC THEN  EXISTS_TAC (--`x:'a`--) THEN ASM_REWRITE_TAC[]
 ]);

	       

val not_IS_EL_EQ_EMPTY = store_thm("not_IS_EL_EQ_EMPTY",
  (--`!(xs:('a)list). (!x. ~(IS_EL x xs)) = (xs = [])`--),
  LIST_INDUCT_TAC
  THEN REWRITE_TAC[theorem "List" "IS_EL",MAP]
  THEN REWRITE_TAC[DE_MORGAN_THM]
  THEN GEN_TAC THEN EQ_TAC  
  THENL
   [REPEAT STRIP_TAC
     THEN XRULE_ASSUM_TAC (ISPEC (--`h:'a`--)) THEN UNDISCH_ONE_TAC THEN REWRITE_TAC[]
   ,
    REWRITE_TAC[GSYM (theorem "List" "NOT_NIL_CONS")]
   ]);

val LIST_EQ = store_thm("LIST_EQ",
  (--`!(xs:('a)list) (ys:('a)list).
    (((LENGTH xs) = (LENGTH ys)) /\ (!n. (n < (LENGTH xs)) ==> ((EL n xs) = (EL n ys)))) ==> (xs = ys)`--),
    
    LIST_INDUCT_TAC
    THENL
     [LIST_INDUCT_TAC
      THENL
       [REPEAT STRIP_TAC THEN REWRITE_TAC[]
       ,
        REWRITE_TAC[theorem "List" "LENGTH"] THEN REWRITE_TAC[GSYM (theorem "num" "NOT_SUC")]
       ]
     ,
      STRIP_TAC THEN LIST_INDUCT_TAC
      THENL
       [REWRITE_TAC[theorem "List" "LENGTH"] THEN REWRITE_TAC[(theorem "num" "NOT_SUC")]
       ,
        REWRITE_TAC[theorem "List" "LENGTH"] THEN REWRITE_TAC[theorem "prim_rec" "INV_SUC_EQ"] 
        THEN REPEAT STRIP_TAC THEN REWRITE_TAC[theorem "list" "CONS_11"]
        THEN SUBGOAL_THEN (--`0 < (SUC (LENGTH (xs:('a)list)))`--) ASSUME_TAC
        THENL[CONV_TAC ARITH_CONV, ALL_TAC]
        THEN RES_TAC
        THEN XRULE_ASSUM_TAC (REWRITE_RULE[definition "list" "EL",definition "list" "HD"])
        THEN ASM_REWRITE_TAC[]
        THEN FIRST_ASSUM MATCH_MP_TAC THEN REPEAT STRIP_TAC 
        THEN IMP_RES_TAC (theorem "prim_rec" "LESS_MONO")
        THEN RES_TAC 
	THEN XRULE_ASSUM_TAC (REWRITE_RULE[definition "list" "EL",definition "list" "TL"])
        THEN ASM_REWRITE_TAC[]
       ]
     ]);


val LIST_EQ_thm = store_thm("LIST_EQ_thm",
  (--`!(xs:('a)list) (ys:('a)list).
    (xs = ys) = (((LENGTH xs) = (LENGTH ys)) /\ (!n. (n < (LENGTH xs)) ==> ((EL n xs) = (EL n ys))))`--),
    
REPEAT GEN_TAC THEN EQ_TAC 
THENL
 [STRIP_TAC THEN ASM_REWRITE_TAC[]
 ,
  MATCH_ACCEPT_TAC LIST_EQ
 ]);
 
    



val EVERY_IS_EL = store_thm("EVERY_IS_EL",
   (--`!(xs:('a)list) (P:'a->bool).
           (EVERY P xs) = (!x. (IS_EL x xs) ==> (P x))`--),
	 
  LIST_INDUCT_TAC THEN REWRITE_TAC[definition "list" "EVERY_DEF",theorem "List" "IS_EL"]
  THEN ASM_REWRITE_TAC[] THEN REPEAT GEN_TAC THEN EQ_TAC THEN REPEAT STRIP_TAC
  THEN ASM_REWRITE_TAC[] THEN FIRST_ASSUM MATCH_MP_TAC  THEN ASM_REWRITE_TAC[]
  );
   

val EVERY_MAP_o = store_thm("EVERY_MAP_o",
  (--`!(xs:('a)list) (f:'b->bool) (g:'a->'b). 
      (EVERY f (MAP g xs)) = (EVERY (f o g) xs)`--),
LIST_INDUCT_TAC THEN REWRITE_TAC[definition "list" "EVERY_DEF",theorem "List" "IS_EL",MAP]
THEN REPEAT GEN_TAC THEN EQ_TAC THEN REPEAT STRIP_TAC
THENL
 [REWRITE_TAC[definition "combin" "o_DEF"] THEN BETA_TAC THEN ASM_REWRITE_TAC[]
 ,XRULE_ASSUM_TAC GSYM THEN ASM_REWRITE_TAC[]
 ,XRULE_ASSUM_TAC (BETA_RULE o (REWRITE_RULE [definition "combin" "o_DEF"])) THEN ASM_REWRITE_TAC[]
 ,ASM_REWRITE_TAC[]
 ]);
 

val IS_EL_IMP_EL = store_thm("IS_EL_IMP_EL",
  (--`!(h:'a) (t:('a)list).
     (IS_EL h t) ==> (?n. (n < (LENGTH t)) /\ (h = (EL n t)))`--),
   GEN_TAC THEN LIST_INDUCT_TAC
   THEN REWRITE_TAC[theorem "List" "IS_EL"]
   THEN REPEAT STRIP_TAC
   THENL
   [EXISTS_TAC (--`0`--) 
     THEN ASM_REWRITE_TAC[theorem "List" "LENGTH",definition "list" "HD", definition "list" "EL"]
     THEN CONV_TAC ARITH_CONV
   ,
    RES_TAC 
    THEN EXISTS_TAC (--`SUC n`--)
    THEN ASM_REWRITE_TAC[theorem "List" "LENGTH",definition "list" "TL", definition "list" "EL"]
    THEN POP_ASSUM (fn thm => ALL_TAC) THEN UNDISCH_ONE_TAC THEN CONV_TAC ARITH_CONV
   ]);


val IS_EL_IMP_least_EL = store_thm("IS_EL_IMP_least_EL",
  (--`!(t:('a)list)(h:'a) .
     (IS_EL h t) ==> (?n. (n < (LENGTH t)) /\ (h = (EL n t)) /\
                     (!k. (k < n) ==> ~(h = (EL k t))))`--),
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[theorem "List" "IS_EL"]
   THEN REPEAT STRIP_TAC
   THENL
   [EXISTS_TAC (--`0`--) 
     THEN ASM_REWRITE_TAC[theorem "List" "LENGTH",definition "list" "HD", definition "list" "EL"]
     THEN REWRITE_TAC[theorem "prim_rec" "NOT_LESS_0"] THEN CONV_TAC ARITH_CONV
   ,
    ASM_CASES_TAC (--`(h:'a) = h'`--)
    THENL
    [ASM_REWRITE_TAC[] THEN EXISTS_TAC (--`0`--) 
     THEN ASM_REWRITE_TAC[theorem "prim_rec" "NOT_LESS_0",EL,HD,LENGTH]
     THEN  CONV_TAC ARITH_CONV
    ,ALL_TAC
    ]
    THEN RES_TAC 
    THEN EXISTS_TAC (--`SUC n`--)
    THEN REWRITE_TAC[theorem "List" "LENGTH",definition "list" "TL", definition "list" "EL"]
    THEN REPEAT CONJ_TAC
    THENL
     [UNDISCH_TAC (--`n < LENGTH (t:('a)list)`--) THEN CONV_TAC ARITH_CONV
     ,ASM_REWRITE_TAC[]
     ,ALL_TAC
     ]
    THEN REPEAT STRIP_TAC
    THEN SUBGOAL_THEN (--`(0 < k) /\ (?m. k = (SUC m))`--) STRIP_ASSUME_TAC
    THENL
     [SUBGOAL_THEN (--`~(k = 0)`--)  ASSUME_TAC  
     THENL[ STRIP_TAC THEN UNDISCH_TAC (--`h' = EL k (CONS (h:'a) t)`--) THEN ASM_REWRITE_TAC[EL,HD]
     THEN STRIP_TAC THEN UNDISCH_TAC (--`~((h:'a) = h')`--) THEN ASM_REWRITE_TAC[],ALL_TAC]
     THEN CONJ_TAC THENL[UNDISCH_ONE_TAC THEN CONV_TAC ARITH_CONV,ALL_TAC]
     THEN POP_ASSUM (fn thm => ASSUME_TAC (REWRITE_RULE[thm] (ISPEC (--`k:num`--) 
              (theorem "arithmetic" "num_CASES")))) THEN ASM_REWRITE_TAC[]
     ,ALL_TAC]
    THEN IMP_RES_TAC EL_CONS THEN POP_ASSUM (fn thm => REWRITE_ASM_TAC[thm])
    THEN UNDISCH_TAC (--`(h':('a)) = EL (PRE k) t`--) THEN REWRITE_TAC[]
    THEN FIRST_ASSUM MATCH_MP_TAC 
    THEN ASM_REWRITE_TAC[theorem "prim_rec" "PRE"]
    THEN UNDISCH_ONE_TAC
    THEN UNDISCH_ONE_TAC THEN UNDISCH_ONE_TAC
    THEN CONV_TAC ARITH_CONV
   ]);




val EL_IMP_IS_EL = store_thm("EL_IMP_IS_EL",
  (--`!(h:'a) (t:('a)list).
      (?n. (n < (LENGTH t)) /\ (h = (EL n t))) ==> (IS_EL h t)`--),
      REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[] THEN MATCH_MP_TAC (theorem "List" "EL_IS_EL")
      THEN ASM_REWRITE_TAC[]);
 
 
 
val IS_EL_EQ_EL = store_thm("IS_EL_EQ_EL",
 (--`!(h:'a) (t:('a)list).
     (IS_EL h t) = (?n. (n < (LENGTH t)) /\ (h = (EL n t)))`--),
   REPEAT GEN_TAC THEN EQ_TAC
   THENL
    [MATCH_ACCEPT_TAC IS_EL_IMP_EL
    ,
     MATCH_ACCEPT_TAC EL_IMP_IS_EL
    ]);
        

val MAP_ID = store_thm("MAP_ID",
  (--`!(xs:('a)list).
     (MAP (\x. x) xs) = xs`--),
    LIST_INDUCT_TAC THEN REWRITE_TAC[MAP]
    THEN BETA_TAC THEN ASM_REWRITE_TAC[]);


val IS_EL_FLAT = store_thm("IS_EL_FLAT",
  (--`!(xss:(('a)list)list) (x:'a) .
             (IS_EL x (FLAT xss)) =  (?ys. (IS_EL ys xss) /\ (IS_EL x ys))`--),
  LIST_INDUCT_TAC 
  THEN REWRITE_TAC[theorem "List" "FLAT", theorem "List" "IS_EL",theorem "List" "IS_EL_APPEND"]
  THEN ASM_REWRITE_TAC[] THEN REPEAT GEN_TAC THEN EQ_TAC THEN STRIP_TAC
  THENL
   [EXISTS_TAC (--`h:('a)list`--) THEN ASM_REWRITE_TAC[]
   ,EXISTS_TAC (--`ys:('a)list`--) THEN ASM_REWRITE_TAC[]
   ,DISJ1_TAC THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[]
   ,DISJ2_TAC THEN EXISTS_TAC (--`ys:('a)list`--) THEN ASM_REWRITE_TAC[]
   ]);
      

add_theory_to_sml "List";

val LENGTH_TL = store_thm("LENGTH_TL",
  (--`!(l:('a)list). (~(l=[])) ==> ((LENGTH l) = (SUC (LENGTH (TL l))))`--),
 LIST_INDUCT_TAC THEN REWRITE_TAC[]
 THEN REPEAT STRIP_TAC THEN REWRITE_TAC[TL,LENGTH]);


val LENGTH_SUC_NOT_NIL = store_thm("LENGTH_SUC_NOT_NIL",
  (--`!(l:('a)list) (n:num).
    ((LENGTH l) = (SUC n)) ==> (~(l=[]))`--),
  LIST_INDUCT_TAC THEN REWRITE_TAC[LENGTH,GSYM (theorem "num" "NOT_SUC")]
  THEN REPEAT GEN_TAC THEN STRIP_TAC THEN REWRITE_TAC[GSYM NOT_NIL_CONS]);

val from = new_recursive_definition
        {name = "from",
         fixity = Prefix,
         rec_axiom = (theorem "prim_rec" "num_Axiom"),
         def = (--`(from k 0 = [])
                   /\
                   (from k (SUC n) = CONS k (from (SUC k) n))`--)};
                   
val IS_EL_from = store_thm ("IS_EL_from",
   (--`!(m:num) (k:num) (n:num) . (IS_EL n (from k m)) ==> ((k <= n) /\ (n < (k + m)))`--),
   
  INDUCT_TAC
  THENL
   [REWRITE_TAC[theorem "arithmetic" "ADD_0", theorem "List" "IS_EL",from]
    THEN CONV_TAC ARITH_CONV
   ,
    REWRITE_TAC[theorem "List" "IS_EL",from]
    THEN REPEAT STRIP_TAC THEN ((ASM_REWRITE_TAC[] THEN CONV_TAC ARITH_CONV) ORELSE RES_TAC)
    THENL[UNDISCH_TAC (--`SUC k <= n`--) THEN CONV_TAC ARITH_CONV,
          UNDISCH_ONE_TAC THEN CONV_TAC ARITH_CONV]
   ]);

val IS_EL_from2 = store_thm ("IS_EL_from2",
   (--`!(m:num) (k:num) (n:num) . ((k <= n) /\ (n < (k + m))) ==> (IS_EL n (from k m))`--),
   
  INDUCT_TAC
  THENL
   [REWRITE_TAC[theorem "arithmetic" "ADD_0", theorem "List" "IS_EL",from]
    THEN CONV_TAC ARITH_CONV
   ,
    REWRITE_TAC[theorem "List" "IS_EL",from]
    THEN REPEAT STRIP_TAC THEN ASM_CASES_TAC (--`(n:num) = k`--) THEN ASM_REWRITE_TAC[]
    THEN FIRST_ASSUM MATCH_MP_TAC THEN UNDISCH_ONE_TAC  THEN UNDISCH_ONE_TAC  THEN UNDISCH_ONE_TAC 
    THEN CONV_TAC ARITH_CONV
   ]);


val IS_EL_from_EQ = store_thm("IS_EL_from_EQ",

  (--`!(m:num) (k:num) (n:num) . (IS_EL n (from k m)) = ((k <= n) /\ (n < (k + m)))`--),
  REPEAT STRIP_TAC THEN EQ_TAC THEN STRIP_TAC
  THEN IMP_RES_TAC IS_EL_from THEN IMP_RES_TAC IS_EL_from2 THEN ASM_REWRITE_TAC[]
);
   


   
val from_LENGTH = store_thm("from_LENGTH",
  (--`!(m:num) (k:num). (LENGTH (from k m)) = m`--),
  INDUCT_TAC
  THEN REWRITE_TAC[from, LENGTH]
  THEN REWRITE_TAC[theorem "prim_rec" "INV_SUC_EQ"]
  THEN ASM_REWRITE_TAC[]);
  
val ADD_CLAUSES = theorem "arithmetic" "ADD_CLAUSES";

val HD_from = store_thm("HD_from",
   (--`!m k. (0 < m) ==> ((HD (from k m)) = k)`--),
   INDUCT_TAC THEN REWRITE_TAC[theorem "prim_rec" "NOT_LESS_0"]
   THEN REPEAT STRIP_TAC THEN REWRITE_TAC[from,HD]);

val TL_from = store_thm("TL_from",
  (--`!m k. (0 < m) ==> ((TL (from k m)) = (from (SUC k) (PRE m)))`--),
  INDUCT_TAC THEN REWRITE_TAC[theorem "prim_rec" "NOT_LESS_0"]
  THEN REPEAT STRIP_TAC THEN REWRITE_TAC[from,TL,theorem "prim_rec" "PRE"]);
 
val EL_from = store_thm("EL_from",
   (--`!n m k. ((0 < m) /\ (n < m)) ==> ((EL n (from k m)) = (k + n))`--),
   INDUCT_TAC
   THENL
    [REPEAT STRIP_TAC THEN REWRITE_TAC[EL,theorem "arithmetic" "ADD_CLAUSES"]
     THEN MATCH_MP_TAC HD_from THEN ASM_REWRITE_TAC[]
    ,
     REPEAT STRIP_TAC THEN REWRITE_TAC[EL] THEN IMP_RES_TAC TL_from
     THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[theorem "arithmetic" "ADD_CLAUSES"]
     THEN ONCE_REWRITE_TAC[theorem "arithmetic" "ADD_SYM"]
     THEN REWRITE_TAC[SYM (el 4 (CONJUNCTS ADD_CLAUSES))]
     THEN ONCE_REWRITE_TAC[theorem "arithmetic" "ADD_SYM"]
     THEN FIRST_ASSUM MATCH_MP_TAC THEN STRIP_ASSUME_TAC (ISPEC (--`m:num`--) (theorem "arithmetic" "num_CASES"))
     THENL
     [UNDISCH_TAC (--`0 < m`--) THEN ASM_REWRITE_TAC[theorem "prim_rec" "LESS_REFL"]
     ,
      ASM_REWRITE_TAC[theorem "prim_rec" "PRE"] THEN UNDISCH_TAC (--`0 < m`--)
      THEN UNDISCH_TAC (--`SUC n < m`--) THEN UNDISCH_TAC (--`m = SUC n'`--)
      THEN CONV_TAC ARITH_CONV
     ]
    ]);
   
    
   
val APPEND_EQ = store_thm("APPEND_EQ",
  (--`!(xs:('a)list) (ys:('a)list) (zs:('a)list) (vs:('a)list).
     ((xs = zs) /\ (ys = vs)) ==> ((APPEND xs ys) = (APPEND zs vs))`--),
    REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[]);
     

val HD_MAP = store_thm("HD_MAP",
  (--`!(xs:('a)list) (f:'a->'b). (0 < (LENGTH xs)) ==> ((HD (MAP f xs)) = (f (HD xs)))`--),
  
   LIST_INDUCT_TAC THEN REWRITE_TAC[LENGTH,theorem "prim_rec" "LESS_REFL"]
   THEN REPEAT STRIP_TAC THEN REWRITE_TAC[MAP,HD]);
   

val TL_MAP = store_thm("TL_MAP",
  (--`!(xs:('a)list) (f:'a->'b). (0 < (LENGTH xs)) ==> ((TL (MAP f xs)) = ((MAP f (TL xs))))`--),
  
   LIST_INDUCT_TAC THEN REWRITE_TAC[LENGTH,theorem "prim_rec" "LESS_REFL"]
   THEN REPEAT STRIP_TAC THEN REWRITE_TAC[MAP,TL]);

val NOT_SUC_LESS_0 = prove(
   (--`!n. ~(SUC n < 0)`--),
   CONV_TAC ARITH_CONV);


fun REWRITE_ASM_TAC thms = 
     XRULE_ASSUM_TAC (BETA_RULE o (REWRITE_RULE thms));
     
    
val MAP_interval = store_thm("MAP_interval",

  (--`!k m (f:num->'a) (g:num->'a). 
         (!n. ((k <= n) /\ (n < (k + m))) ==> ((f n) = (g n))) ==> ((MAP f (from k m)) = (MAP g (from k m)))`--),

    REPEAT STRIP_TAC THEN MATCH_MP_TAC LIST_EQ
    THEN CONJ_TAC
    THENL
     [REWRITE_TAC[LENGTH_MAP]
     ,
      STRIP_TAC 
      THEN UNDISCH_ONE_TAC THEN SPEC_TAC ((--`k:num`--), (--`k:num`--))
      THEN SPEC_TAC ((--`m:num`--), (--`m:num`--))
      THEN SPEC_TAC ((--`n:num`--), (--`n:num`--))
      THEN INDUCT_TAC
      THENL
       [REWRITE_TAC[LENGTH_MAP] THEN REPEAT STRIP_TAC THEN REWRITE_TAC[EL]
        THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'a`==, residue = ==`:num`==},
                                     {redex = ==`:'b`==, residue = ==`:'a`==}] HD_MAP) THEN ASM_REWRITE_TAC[]
        THEN REWRITE_ASM_TAC[from_LENGTH] THEN STRIP_ASSUME_TAC (ISPEC (--`m:num`--) (theorem "arithmetic" "num_CASES"))
        THENL[UNDISCH_TAC (--`0 < m`--) THEN ASM_REWRITE_TAC[theorem "prim_rec" "LESS_REFL"],ALL_TAC]
        THEN ASM_REWRITE_TAC[from,HD] THEN FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[]
        THEN CONV_TAC ARITH_CONV
       ,
        REWRITE_TAC[LENGTH_MAP] THEN REPEAT STRIP_TAC THEN REWRITE_TAC[EL]
        THEN REWRITE_TAC[TL_MAP] THEN SUBGOAL_THEN (--`0 < (LENGTH (from (k:num) (m:num)))`--) ASSUME_TAC
        THENL[UNDISCH_ONE_TAC THEN REWRITE_TAC[from_LENGTH] THEN CONV_TAC ARITH_CONV,ALL_TAC]
        THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'a`==, residue = ==`:num`==},
                                     {redex = ==`:'b`==, residue = ==`:'a`==}] TL_MAP) THEN ASM_REWRITE_TAC[]
        THEN REWRITE_ASM_TAC[from_LENGTH] THEN STRIP_ASSUME_TAC (ISPEC (--`m:num`--) (theorem "arithmetic" "num_CASES"))
        THENL[UNDISCH_TAC (--`0 < m`--) THEN ASM_REWRITE_TAC[theorem "prim_rec" "LESS_REFL"],ALL_TAC]
        THEN ASM_REWRITE_TAC[from,TL] 
        THEN REWRITE_ASM_TAC[LENGTH_MAP,from_LENGTH]
        THEN SUBGOAL_THEN (--`n < n'`--) ASSUME_TAC
        THENL[UNDISCH_TAC (--`SUC n < m`--) THEN ASM_REWRITE_TAC[] THEN CONV_TAC ARITH_CONV,ALL_TAC]
        THEN SUBGOAL_THEN (--`(!n. (((SUC k) <= n) /\ (n < ((SUC k) + n'))) ==> (((f:num->'a) n) = (g n)))`--) ASSUME_TAC
        THENL[REPEAT STRIP_TAC THEN FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[]
              THEN UNDISCH_ONE_TAC THEN UNDISCH_ONE_TAC THEN CONV_TAC ARITH_CONV,ALL_TAC]
        THEN RES_TAC
       ]
     ]);
     


val EL_MAP_from = store_thm("EL_MAP_from", 
  
  (--`!m  k  n (f:num->'b). ((0 < m) /\ (n < m)) ==> ((EL n (MAP f (from k m))) = (f (k + n)))`--), 
   
   REPEAT STRIP_TAC
   THEN SUBGOAL_THEN (--`n < (LENGTH (from k m))`--) ASSUME_TAC
   THENL[ASM_REWRITE_TAC[from_LENGTH],ALL_TAC] 
   THEN IMP_RES_TAC (INST_TYPE [{redex = ==`:'a`==, residue = ==`:num`==}] EL_MAP)
   THEN ASM_REWRITE_TAC[] THEN IMP_RES_TAC EL_from THEN ASM_REWRITE_TAC[]);

val splid = new_definition ("splid",
   (--`splid (p:'a->'b) (x:'a) = (p x, x)`--));
   

val SND_splid = store_thm("SND_splid",
  (--`!(f:'a->'b). (SND o (splid f)) = (\x. x)`--),
    STRIP_TAC THEN CONV_TAC FUN_EQ_CONV
    THEN REWRITE_TAC[definition "combin" "o_DEF"]
    THEN BETA_TAC THEN REWRITE_TAC[splid]);

val FST_splid = store_thm("FST_splid",
   (--`!(f:'a->'b). (FST o (splid f)) = f`--),    
    STRIP_TAC THEN CONV_TAC FUN_EQ_CONV
    THEN REWRITE_TAC[definition "combin" "o_DEF"]
    THEN BETA_TAC THEN REWRITE_TAC[splid]);


(* properties of ZIP, MAP, splid *)


val ZIP_MAP_EQ_MAP_splid = store_thm("ZIP_MAP_EQ_MAP_splid",
  (--`!(f:'a->'b) (xs:('a)list).
      (ZIP ((MAP f xs), xs)) = (MAP (splid f) xs)`--),
      
   STRIP_TAC THEN LIST_INDUCT_TAC
   THENL
    [REWRITE_TAC[definition "list" "MAP",definition "List" "ZIP"]
    ,
     ASM_REWRITE_TAC[definition "list" "MAP",definition "List" "ZIP",splid]
    ]);


val ZIP = definition "List" "ZIP";

val HD_ZIP = store_thm("HD_ZIP",
  (--`!(xs:('a)list) (ys:('b)list).
      ((0 < (LENGTH xs)) /\ (LENGTH xs = LENGTH ys)) ==> ((HD (ZIP (xs,ys))) = (HD xs, HD ys))`--),
      
   LIST_INDUCT_TAC THEN REWRITE_TAC[LENGTH,theorem "prim_rec" "LESS_REFL"]
   THEN CONV_TAC (ONCE_DEPTH_CONV SYM_CONV) THEN REWRITE_TAC[LENGTH_CONS]
   THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[ZIP,HD]);
   

val TL_ZIP = store_thm("TL_ZIP  ",
  (--`!(xs:('a)list) (ys:('b)list).
      ((0 < (LENGTH xs)) /\ (LENGTH xs = LENGTH ys)) ==> ((TL (ZIP (xs,ys))) = (ZIP (TL xs, TL ys)))`--),
      
    LIST_INDUCT_TAC THEN REWRITE_TAC[LENGTH,theorem "prim_rec" "LESS_REFL"]
    THEN CONV_TAC (ONCE_DEPTH_CONV SYM_CONV) THEN REWRITE_TAC[LENGTH_CONS] 
    THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[ZIP,TL]);

 
val EL_ZIP = store_thm("EL_ZIP",
  (--`!(n:num) (xs:('a)list) (ys:('b)list).
     ((n < (LENGTH xs)) /\ (LENGTH xs = LENGTH ys)) ==> ((EL n (ZIP (xs,ys))) = (EL n xs, EL n ys))`--),
 
   INDUCT_TAC
   THENL
    [REPEAT STRIP_TAC THEN REWRITE_TAC[EL] THEN MATCH_MP_TAC HD_ZIP
     THEN ASM_REWRITE_TAC[]
    ,
     REPEAT STRIP_TAC THEN REWRITE_TAC[EL]
     THEN SUBGOAL_THEN (--`0 < (LENGTH (xs:('a)list))`--) ASSUME_TAC
     THENL[UNDISCH_TAC (--`SUC n < LENGTH (xs:('a)list)`--) THEN CONV_TAC ARITH_CONV,ALL_TAC]
     THEN IMP_RES_TAC TL_ZIP THEN ASM_REWRITE_TAC[] THEN FIRST_ASSUM MATCH_MP_TAC
     THEN STRIP_ASSUME_TAC (ISPEC (--`ys:('b)list`--) list_CASES)
     THENL[UNDISCH_TAC (--`0 < LENGTH (xs:('a)list)`--) THEN ASM_REWRITE_TAC[LENGTH] 
           THEN REWRITE_TAC[theorem "prim_rec" "LESS_REFL"],ALL_TAC]
     THEN STRIP_ASSUME_TAC (ISPEC (--`xs:('a)list`--) list_CASES)
     THENL[UNDISCH_TAC  (--`LENGTH (xs:('a)list) = LENGTH (ys:('b)list)`--) THEN ASM_REWRITE_TAC[LENGTH] 
           THEN REWRITE_TAC[GSYM(theorem "num" "NOT_SUC")],ALL_TAC]
     THEN ASM_REWRITE_TAC[TL] 
     THEN UNDISCH_TAC (--`LENGTH (xs:('a)list) = LENGTH (ys:('b)list)`--) 
     THEN UNDISCH_TAC (--`SUC n < LENGTH (xs:('a)list)`--)
     THEN ASM_REWRITE_TAC[LENGTH] THEN CONV_TAC ARITH_CONV
    ]);



val LENGTH_ZIP_L = store_thm("LENGTH_ZIP_L",
 (--`!(l1:('a)list) (l2:('b)list).
       (LENGTH l1 = LENGTH l2) ==>
       (LENGTH (ZIP (l1,l2)) = LENGTH l1)`--),
   REPEAT STRIP_TAC THEN IMP_RES_TAC (theorem "List" "LENGTH_ZIP"));
       
val LENGTH_ZIP_R = store_thm("LENGTH_ZIP_R",
 (--`!(l1:('a)list) (l2:('b)list).
       (LENGTH l1 = LENGTH l2) ==>
       (LENGTH (ZIP (l1,l2)) = LENGTH l2)`--),
   REPEAT STRIP_TAC THEN IMP_RES_TAC (theorem "List" "LENGTH_ZIP"));
   

val ZIP_MAP_LENGTH = store_thm("ZIP_MAP_LENGTH",
  (--`!(xs:('a)list) (f:'a->'b) (g:'a->'c).
    (LENGTH (ZIP((MAP f xs), (MAP g xs)))) = (LENGTH (MAP f xs))`--),
    
   REPEAT GEN_TAC THEN MATCH_MP_TAC  LENGTH_ZIP_L
   THEN REWRITE_TAC[theorem "List" "LENGTH_MAP"]);
   
val ZIP_MAP_LENGTH_R = store_thm("ZIP_MAP_LENGTH_R",
  (--`!(xs:('a)list) (f:'a->'b) (g:'a->'c).
    (LENGTH (ZIP((MAP f xs), (MAP g xs)))) = (LENGTH (MAP g xs))`--),
    
   REPEAT GEN_TAC THEN MATCH_MP_TAC  LENGTH_ZIP_R
   THEN REWRITE_TAC[theorem "List" "LENGTH_MAP"]);
  

val ZIP_EQ = store_thm("ZIP_EQ",
  (--`!(xs1:('a)list) (ys1:('b)list) (xs2:('a)list) (ys2:('b)list).
   ((xs1 = xs2) /\ (ys1 = ys2)) ==> ((ZIP(xs1,ys1))=(ZIP(xs2,ys2)))`--),
   REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[]);
   
val FOLDR_EQ = store_thm("FOLDR_EQ",
   (--`!(xs:('a)list) (ys:('a)list) (f:'a->'b->'b) (e:'b).
     (xs = ys) ==> ((FOLDR f e xs) = (FOLDR f e ys))`--),
   REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[]);  


val MAP_EQ = store_thm("MAP_EQ",
   (--`!(xs:('a)list) (f:'a->'b) (g:'a->'b).
     (!x. (IS_EL x xs) ==> ((f x) = (g x))) ==> ((MAP f xs) = (MAP g xs))`--),
  LIST_INDUCT_TAC 
  THEN REWRITE_TAC[definition "list" "MAP", theorem "list" "CONS_11",theorem "List" "IS_EL"]
  THEN REPEAT STRIP_TAC THEN FIRST_ASSUM MATCH_MP_TAC THEN REWRITE_TAC[]
  THEN REPEAT STRIP_TAC THEN RES_TAC);  



val DEL = new_recursive_definition
    {name = "DEL",
         fixity = Prefix,
         rec_axiom = (theorem "list" "list_Axiom"),
         def = (--`(DEL y ([]:('a)list) = ([]:('a)list))
                   /\
                   (DEL y (CONS x xs) = ((x = y) => xs | (CONS x (DEL y xs))))`--)};



val DEL_APPEND = store_thm("DEL_APPEND",
  (--`!(xs:('a)list) (ys:('a)list) (x:'a).
     (~(IS_EL x xs)) ==>
      ((DEL x (APPEND (APPEND xs [x]) ys)) = (APPEND  xs ys))`--),
      
LIST_INDUCT_TAC
THENL
 [REPEAT STRIP_TAC THEN REWRITE_TAC[APPEND_NIL,GSYM CONS_APPEND,DEL]
 ,
  REWRITE_TAC[IS_EL,DE_MORGAN_THM,GSYM APPEND_ASSOC]
  THEN REWRITE_TAC[APPEND] THEN REPEAT STRIP_TAC 
  THEN UNDISCH_TAC (--`~((x:'a) = h)`--) THEN STRIP_TAC
  THEN POP_ASSUM (fn thm => ASSUME_TAC (GSYM thm))
  THEN ASM_REWRITE_TAC[DEL] THEN REWRITE_TAC[CONS_11]
  THEN ONCE_REWRITE_TAC[CONS_APPEND] THEN REWRITE_TAC[APPEND_ASSOC] THEN RES_TAC
  THEN ASM_REWRITE_TAC[]
 ]  );
   


val NO_DUPLICATES = new_definition("NO_DUPLICATES",
   (--`NO_DUPLICATES (xs:('a)list) = 
       (!n k. ((n < (LENGTH xs)) /\ (k < (LENGTH xs)) /\ (~(n = k))) 
           ==> (~((EL n xs) = (EL k xs))))`--));



val NO_DUPLICATES_CONS = store_thm("NO_DUPLICATES_CONS",
  (--`!(xs:('a)list) (h:'a).
       (NO_DUPLICATES (CONS h xs)) ==> (NO_DUPLICATES xs)`--),

REWRITE_TAC[NO_DUPLICATES,LENGTH] THEN REPEAT GEN_TAC THEN STRIP_TAC
THEN REPEAT GEN_TAC THEN STRIP_TAC
THEN SUBGOAL_THEN (--`~(NULL (xs:('a)list))`--) ASSUME_TAC
THENL[REWRITE_TAC[NULL_EQ_NIL] THEN STRIP_TAC THEN UNDISCH_TAC (--`n < LENGTH (xs:('a)list)`--) 
THEN ASM_REWRITE_TAC[LENGTH,theorem "prim_rec" "NOT_LESS_0"],ALL_TAC]
THEN REWRITE_ASM_TAC[GSYM LENGTH_NOT_NULL]
THEN XRULE_ASSUM_TAC (ISPEC (--`(SUC n):num`--)) 
THEN XRULE_ASSUM_TAC (ISPEC (--`(SUC k):num`--)) 
THEN REWRITE_ASM_TAC[EL,TL] THEN FIRST_ASSUM MATCH_MP_TAC
THEN DO_TAC UNDISCH_ONE_TAC 4 THEN CONV_TAC ARITH_CONV );



val NO_DUPLICATES_THM = store_thm("NO_DUPLICATES_THM",
  (--`!(xs:('a)list). (NO_DUPLICATES xs) ==> (!(x:'a). (IS_EL x xs) ==> (~(IS_EL x (DEL x xs))))`--),
  
LIST_INDUCT_TAC 
THENL[REWRITE_TAC[NO_DUPLICATES,LENGTH,theorem "prim_rec" "NOT_LESS_0",IS_EL]  ,ALL_TAC]
THEN GEN_TAC THEN STRIP_TAC THEN GEN_TAC THEN STRIP_TAC
THEN REWRITE_TAC[DEL] THEN COND_CASES_TAC
THENL
 [ALL_TAC
 ,
  REWRITE_TAC[IS_EL,DE_MORGAN_THM] THEN CONV_TAC (ONCE_DEPTH_CONV SYM_CONV) THEN ASM_REWRITE_TAC[]
  THEN UNDISCH_TAC (--`IS_EL (x:'a) (CONS h xs)`--) THEN REWRITE_TAC[IS_EL]
  THEN CONV_TAC (ONCE_DEPTH_CONV SYM_CONV) THEN ASM_REWRITE_TAC[]
  THEN STRIP_TAC THEN IMP_RES_TAC NO_DUPLICATES_CONS THEN RES_TAC
 ]
THEN POP_ASSUM (fn thm => REWRITE_TAC[SYM thm] THEN REWRITE_ASM_TAC[SYM thm])
THEN POP_ASSUM (fn thm => ALL_TAC)
THEN REWRITE_TAC[IS_EL_EQ_EL ]
THEN CONV_TAC NOT_EXISTS_CONV THEN REWRITE_TAC[DE_MORGAN_THM]
THEN REWRITE_TAC[GSYM IMP_DISJ_THM] THEN UNDISCH_ONE_TAC THEN REWRITE_TAC[NO_DUPLICATES]
THEN STRIP_TAC THEN GEN_TAC THEN STRIP_TAC
THEN XRULE_ASSUM_TAC (ISPEC (--`SUC n`--))
THEN XRULE_ASSUM_TAC (ISPEC (--`0`--))
THEN REWRITE_ASM_TAC[LENGTH,EL,HD,TL]
THEN CONV_TAC (ONCE_DEPTH_CONV SYM_CONV) THEN FIRST_ASSUM MATCH_MP_TAC
THEN UNDISCH_ONE_TAC THEN CONV_TAC ARITH_CONV

);


val split_LIST = store_thm("split_LIST",

  (--`!(xs:('a)list) (x:'a).
     (IS_EL x xs) ==>
     (?ys zs. (xs = (APPEND (APPEND ys [x]) zs)) /\ ((DEL x xs) = (APPEND ys zs)))`--),
     
     
REPEAT STRIP_TAC THEN IMP_RES_TAC IS_EL_IMP_least_EL
THEN SUBGOAL_THEN (--`~(NULL (xs:('a)list))`--) ASSUME_TAC
THENL[REWRITE_TAC[NULL_EQ_NIL] THEN STRIP_TAC THEN UNDISCH_TAC (--`n < LENGTH (xs:('a)list)`--) 
THEN ASM_REWRITE_TAC[LENGTH,theorem "prim_rec" "NOT_LESS_0"],ALL_TAC]
THEN ASM_CASES_TAC (--`n = 0`--)
THENL
 [POP_ASSUM (fn thm => REWRITE_ASM_TAC[thm,EL])
 THEN EXISTS_TAC (--`[]:('a)list`--) THEN EXISTS_TAC (--`TL (xs:('a)list)`--)
 THEN REWRITE_TAC[APPEND_NIL] THEN IMP_RES_TAC CONS
 THEN POP_ASSUM (fn thm => ONCE_REWRITE_TAC[SYM thm])
 THEN XRULE_ASSUM_TAC SYM THEN ASM_REWRITE_TAC[HD,TL,DEL]
 THEN MATCH_ACCEPT_TAC CONS_APPEND
 ,ALL_TAC
 ]
THEN EXISTS_TAC (--`MAP (\n. (EL n (xs:('a)list))) (from 0 n)`--)
THEN EXISTS_TAC (--`MAP (\n. (EL n (xs:('a)list))) (from (SUC n) ((LENGTH xs) - (SUC n)))`--)
THEN SUBGOAL_THEN (--`((xs:('a)list) =
       APPEND (APPEND (MAP (\n. EL n xs) (from 0 n)) [x])
         (MAP (\n. EL n xs) (from (SUC n) (LENGTH xs - SUC n))))`--) ASSUME_TAC
THENL
 [MATCH_MP_TAC LIST_EQ
  THEN REWRITE_TAC[LENGTH_APPEND,LENGTH_MAP,from_LENGTH,LENGTH]
  THEN CONJ_TAC
  THENL
   [ONCE_REWRITE_TAC[theorem "arithmetic" "ADD_CLAUSES"]
    THEN REWRITE_TAC[theorem "arithmetic" "ADD_0"] 
    THEN UNDISCH_TAC (--`n < LENGTH (xs:('a)list)`--) THEN CONV_TAC ARITH_CONV
   ,ALL_TAC]
  THEN REPEAT STRIP_TAC
  THEN ASM_CASES_TAC (--`n' = (n:num)`--)
  THENL
   [ASM_REWRITE_TAC[]
    THEN SUBGOAL_THEN (--`n < (LENGTH (APPEND (MAP (\n. EL n (xs:('a)list)) (from 0 n)) [EL n xs]))`--)
      (fn thm => REWRITE_TAC[ (REWRITE_RULE [thm] 
                  (ISPEC (--`(APPEND (MAP (\n. EL n (xs:('a)list)) (from 0 n)) [EL n xs])`--)
                     (ISPEC (--`n:num`--) EL_APPEND1)))])
    THENL[REWRITE_TAC[LENGTH_APPEND,LENGTH_MAP,from_LENGTH,LENGTH] THEN CONV_TAC ARITH_CONV,ALL_TAC]
    THEN SUBGOAL_THEN (--`(LENGTH (MAP (\n. EL n (xs:('a)list)) (from 0 n))) <= n`--)
      (fn thm => REWRITE_TAC[ (REWRITE_RULE [thm] (ISPEC (--`n:num`--) 
                  (ISPEC (--`(MAP (\n. EL n (xs:('a)list)) (from 0 n))`--) EL_APPEND2)))])
    THENL[REWRITE_TAC[LENGTH_MAP,from_LENGTH,LENGTH] THEN CONV_TAC ARITH_CONV,ALL_TAC]
    THEN REWRITE_TAC[LENGTH_MAP,from_LENGTH,LENGTH,theorem "arithmetic" "SUB_EQUAL_0",EL,HD] 
   ,ALL_TAC]
  THEN ASM_CASES_TAC (--`n' < (n:num)`--)
  THENL
   [SUBGOAL_THEN (--`n' < (LENGTH (APPEND (MAP (\n. EL n (xs:('a)list)) (from 0 n)) [x]))`--)
      (fn thm => REWRITE_TAC[ (REWRITE_RULE [thm] 
                  (ISPEC (--`(APPEND (MAP (\n. EL n (xs:('a)list)) (from 0 n)) [x])`--)
                     (ISPEC (--`n':num`--) EL_APPEND1)))])
    THENL[REWRITE_TAC[LENGTH_APPEND,LENGTH_MAP,from_LENGTH,LENGTH] 
          THEN UNDISCH_ONE_TAC THEN CONV_TAC ARITH_CONV,ALL_TAC]
    THEN SUBGOAL_THEN (--`n' < (LENGTH (MAP (\n. EL n (xs:('a)list)) (from 0 n)))`--)
      (fn thm => REWRITE_TAC[ (REWRITE_RULE [thm] 
                  (ISPEC (--`(MAP (\n. EL n (xs:('a)list)) (from 0 n))`--) 
		  (ISPEC (--`n':num`--) EL_APPEND1)))])
    THENL[REWRITE_TAC[LENGTH_MAP,from_LENGTH,LENGTH] THEN ASM_REWRITE_TAC[],ALL_TAC]
    THEN SUBGOAL_THEN (--`n' < LENGTH (from 0 n)`--)
       (fn thm => REWRITE_TAC [REWRITE_RULE[thm] 
            (ISPEC (--`(from 0 n)`--) (ISPEC (--`n':num`--) 
	     (INST_TYPE [{redex = ==`:'a`==, residue = ==`:num`==}, 
                             {redex = ==`:'b`==, residue = ==`:'a`==}] EL_MAP))) ] )
    THENL[ASM_REWRITE_TAC[from_LENGTH],ALL_TAC]
    THEN BETA_TAC
    THEN SUBGOAL_THEN (--`(0 < n) /\ (n' < n)`--)
        (fn thm => REWRITE_TAC[definition "arithmetic" "ADD", REWRITE_RULE[thm]
	    (ISPEC (--`0`--) (ISPEC (--`n:num`--) (ISPEC (--`n':num`--) EL_from)))])
    THEN ASM_REWRITE_TAC[] THEN UNDISCH_TAC (--`~(n = 0)`--) THEN CONV_TAC ARITH_CONV
   ,
    SUBGOAL_THEN (--`(SUC n) <= n'`--) ASSUME_TAC
    THENL[UNDISCH_ONE_TAC THEN UNDISCH_ONE_TAC THEN CONV_TAC ARITH_CONV,ALL_TAC]
    THEN SUBGOAL_THEN (--`(LENGTH (APPEND (MAP (\n. EL n (xs:('a)list)) (from 0 n)) [x])) <= n'`--)
      (fn thm => REWRITE_TAC[ (REWRITE_RULE [thm] (ISPEC (--`n':num`--) 
                  (ISPEC (--`APPEND (MAP (\n. EL n (xs:('a)list)) (from 0 n)) [x]`--) EL_APPEND2)))])
    THENL[REWRITE_TAC[LENGTH_APPEND,LENGTH_MAP,from_LENGTH,LENGTH] 
          THEN UNDISCH_ONE_TAC THEN CONV_TAC ARITH_CONV,ALL_TAC]
    THEN REWRITE_TAC[LENGTH_APPEND,LENGTH_MAP,from_LENGTH,LENGTH]
    THEN SUBGOAL_THEN (--`(n'  - (n + (SUC 0))) < LENGTH (from (SUC n) ((LENGTH (xs:('a)list)) - (SUC n)))`--)
       (fn thm => REWRITE_TAC [REWRITE_RULE[thm] 
            (ISPEC (--`(from (SUC n) ((LENGTH (xs:('a)list)) - (SUC n)))`--) 
	       (ISPEC (--`n' - (n + (SUC 0))`--) 
	     (INST_TYPE [{redex = ==`:'a`==, residue = ==`:num`==}, 
                             {redex = ==`:'b`==, residue = ==`:'a`==}] EL_MAP))) ] )
    THENL[REWRITE_TAC[from_LENGTH] THEN DO_TAC UNDISCH_ONE_TAC 5
          THEN UNDISCH_TAC (--`n < LENGTH (xs:('a)list)`--) THEN CONV_TAC ARITH_CONV,ALL_TAC]
    THEN BETA_TAC
    THEN SUBGOAL_THEN (--`0 < LENGTH (xs:('a)list) - SUC n /\ n' - (n + SUC 0) < LENGTH xs - SUC n`--)
        (fn thm => REWRITE_TAC[REWRITE_RULE[thm]
	    (ISPEC (--`SUC n`--) (ISPEC (--`(LENGTH (xs:('a)list) - SUC n)`--) 
	     (ISPEC (--`n' - (n + (SUC 0))`--) EL_from)))])
    THENL[DO_TAC UNDISCH_ONE_TAC 5 THEN UNDISCH_TAC (--`n < LENGTH (xs:('a)list)`--) 
          THEN CONV_TAC ARITH_CONV,ALL_TAC]
    THEN SUBGOAL_THEN (--`(SUC n + (n' - (n + SUC 0))) = n'`--)
        (fn thm => REWRITE_TAC[thm])
    THEN DO_TAC UNDISCH_ONE_TAC 3  THEN CONV_TAC ARITH_CONV
   ]
  
 ,
  CONJ_TAC THENL[FIRST_ASSUM MATCH_ACCEPT_TAC, ALL_TAC]
  THEN POP_ASSUM (fn thm => CONV_TAC (RATOR_CONV (ONCE_REWRITE_CONV [thm])))
  THEN MATCH_MP_TAC DEL_APPEND
  THEN REWRITE_TAC[IS_EL_EQ_EL ] THEN CONV_TAC NOT_EXISTS_CONV
  THEN REWRITE_TAC[DE_MORGAN_THM,GSYM IMP_DISJ_THM,LENGTH_MAP,from_LENGTH]
  THEN GEN_TAC THEN STRIP_TAC 
  THEN  SUBGOAL_THEN (--`n' < LENGTH (from 0 n)`--)
       (fn thm => REWRITE_TAC [REWRITE_RULE[thm] 
            (ISPEC (--`(from 0 n)`--) 
	       (ISPEC (--`n':num`--) 
	     (INST_TYPE [{redex = ==`:'a`==, residue = ==`:num`==}, 
                             {redex = ==`:'b`==, residue = ==`:'a`==}] EL_MAP))) ] )
  THENL[ASM_REWRITE_TAC[from_LENGTH],ALL_TAC]
  THEN BETA_TAC THEN FIRST_ASSUM MATCH_MP_TAC
  THEN SUBGOAL_THEN (--`(0 < n) /\ (n' < n)`--)
      (fn thm => ASM_REWRITE_TAC [definition "arithmetic" "ADD",REWRITE_RULE[thm] 
          (ISPEC (--`0`--) (ISPEC (--`n:num`--) (ISPEC (--`n':num`--) EL_from)))])
  THEN ASM_REWRITE_TAC[] THEN UNDISCH_TAC (--`~(n = 0)`--) THEN CONV_TAC ARITH_CONV
	  
 ]);
 

add_theory_to_sml "operator";


val FOLDR_COMM_MONOID_APPEND = prove_thm("FOLDR_COMM_MONOID_APPEND",

  (--`!(op:'a->'a->'a) (e:'a).
     ((MONOID op e) /\ (COMM op)) ==> 
     (!(xs:('a)list) (ys:('a)list). (FOLDR op e (APPEND xs ys)) = (op (FOLDR op e xs) (FOLDR op e ys)))`--),	
GEN_TAC THEN GEN_TAC THEN STRIP_TAC THEN LIST_INDUCT_TAC
THENL
[REWRITE_TAC[APPEND,FOLDR] THEN REWRITE_ASM_TAC[MONOID_DEF,LEFT_ID_DEF]
 THEN ASM_REWRITE_TAC[] 
,
 GEN_TAC THEN LIST_INDUCT_TAC
 THENL[REWRITE_TAC[APPEND_NIL,FOLDR] 
 THEN REWRITE_ASM_TAC[MONOID_DEF,RIGHT_ID_DEF]
 THEN ASM_REWRITE_TAC[] ,ALL_TAC]
 THEN GEN_TAC THEN REWRITE_TAC[APPEND,FOLDR] THEN ASM_REWRITE_TAC[FOLDR]
 THEN REWRITE_ASM_TAC[MONOID_DEF,ASSOC_DEF]
 THEN ASM_REWRITE_TAC[]
]
);

val MONOID_COMM_FOLDR_FILTER = store_thm("MONOID_COMM_FOLDR_FILTER",
 (--`!(op:'a->'a->'a) (e:'a) xs.
   ((MONOID op e) /\ (COMM op)) ==>
    ((FOLDR op e xs) = (FOLDR op e (FILTER (\x. ~(x = e)) xs)))`--),
    
GEN_TAC THEN GEN_TAC THEN LIST_INDUCT_TAC
THEN REWRITE_TAC[FOLDR,FILTER]
THEN BETA_TAC THEN REPEAT STRIP_TAC THEN COND_CASES_TAC
THEN REWRITE_TAC[FOLDR]
THENL
 [RES_TAC THEN ASM_REWRITE_TAC[]
 ,UNDISCH_ONE_TAC THEN REWRITE_TAC[] THEN STRIP_TAC THEN ASM_REWRITE_TAC[] THEN RES_TAC
 THEN REWRITE_ASM_TAC[MONOID_DEF,LEFT_ID_DEF] THEN ASM_REWRITE_TAC[]
 ]);
 



val MONOID_COMM_FOLDR_DEL = prove_thm("MONOID_COMM_FOLDR_DEL",

(--`!(op:'b->'b->'b) (e:'b). 
    ((MONOID op e) /\ (COMM op))
    ==> 
     (!(xs:('a)list) (x:'a) (f:'a->'b). 
       (IS_EL x xs) ==>
          ((FOLDR op e (MAP f xs))
          =
	  (op (FOLDR op e (MAP f (DEL x xs))) (f x))))`--),
	
	
REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[]
THEN IMP_RES_TAC split_LIST THEN ASM_REWRITE_TAC[]
THEN REWRITE_TAC[MAP_APPEND]
THEN DO_TAC (POP_ASSUM (fn thm => ALL_TAC)) 2
THEN IMP_RES_TAC FOLDR_COMM_MONOID_APPEND THEN ASM_REWRITE_TAC[]
THEN REWRITE_TAC[MAP,FOLDR] THEN REWRITE_ASM_TAC[MONOID_DEF,RIGHT_ID_DEF]
THEN ASM_REWRITE_TAC[] THEN UNDISCH_TAC (--`COMM (op:'b->'b->'b)`--)
THEN STRIP_TAC
THEN POP_ASSUM (fn thm => 
   CONV_TAC (RATOR_CONV (RAND_CONV (RATOR_CONV (ONCE_REWRITE_CONV [(REWRITE_RULE[COMM_DEF] thm)]))))
   THEN ASSUME_TAC thm)
THEN POP_ASSUM (fn thm => 
   CONV_TAC (RAND_CONV (ONCE_REWRITE_CONV [(REWRITE_RULE[COMM_DEF] thm)]))) 
THEN REWRITE_ASM_TAC[ASSOC_DEF] THEN ASM_REWRITE_TAC[]);

  
  
val FOLDR_ID = prove_thm("FOLDR_ID",
  (--`!(xs:('a)list) (op:'a->'a->'a) (e:'a).
     ((MONOID op e) /\ (COMM op) /\ (!x. (IS_EL x xs) ==> (x = e)))
     ==>
     (FOLDR op e xs = e)`--),
     
LIST_INDUCT_TAC THEN REWRITE_TAC[FOLDR,IS_EL] 
THEN REPEAT STRIP_TAC 
THEN SUBGOAL_THEN (--`(h:'a) = e`--) (fn thm => REWRITE_TAC[thm])
THENL[FIRST_ASSUM MATCH_MP_TAC THEN REWRITE_TAC[],ALL_TAC]
THEN RES_TAC THEN DO_TAC (POP_ASSUM (fn thm => ALL_TAC)) 2
THEN REWRITE_ASM_TAC[MONOID_DEF,LEFT_ID_DEF] THEN ASM_REWRITE_TAC[]
THEN FIRST_ASSUM MATCH_MP_TAC THEN REPEAT STRIP_TAC THEN RES_TAC);


val ZIP_APPEND = prove_thm("ZIP_APPEND",
   (--`!(lv1:('a)list) (lv2:('a)list) (le1:('b)list) (le2:('b)list).
     ((LENGTH lv1 = LENGTH le1) /\ (LENGTH lv2 = LENGTH le2))
    ==>
    ((ZIP ((APPEND lv1 lv2), (APPEND le1 le2))) = (APPEND (ZIP (lv1, le1)) (ZIP (lv2, le2))))`--),
    
   LIST_INDUCT_TAC
   THEN REWRITE_TAC[LENGTH]
   THENL
    [LIST_INDUCT_TAC THEN REWRITE_TAC[LENGTH,APPEND]
     THENL
      [REPEAT STRIP_TAC THEN XRULE_ASSUM_TAC SYM THEN REWRITE_ASM_TAC[LENGTH_NIL]
       THEN ASM_REWRITE_TAC[APPEND,ZIP]
      ,
       REPEAT STRIP_TAC THEN XRULE_ASSUM_TAC SYM THEN REWRITE_ASM_TAC[LENGTH_NIL,LENGTH_CONS]
       THEN UNDISCH_ONE_TAC THEN STRIP_TAC THEN ASM_REWRITE_TAC[APPEND,ZIP]
      ]
    ,
     GEN_TAC THEN LIST_INDUCT_TAC THEN REWRITE_TAC[LENGTH]
     THENL
      [REPEAT STRIP_TAC THEN XRULE_ASSUM_TAC SYM THEN REWRITE_ASM_TAC[LENGTH_NIL,LENGTH_CONS]
       THEN DO_TAC UNDISCH_ONE_TAC 2 THEN REPEAT STRIP_TAC
       THEN ASM_REWRITE_TAC[APPEND,ZIP,CONS_11,APPEND_NIL]
      ,
       REPEAT STRIP_TAC THEN XRULE_ASSUM_TAC SYM THEN REWRITE_ASM_TAC[LENGTH_CONS]
       THEN DO_TAC UNDISCH_ONE_TAC 2 THEN REPEAT STRIP_TAC
       THEN ASM_REWRITE_TAC[] THEN REWRITE_TAC[APPEND,CONS_11,ZIP]
       THEN REWRITE_TAC[GSYM ZIP] THEN FIRST_ASSUM MATCH_MP_TAC
       THEN ASM_REWRITE_TAC[LENGTH]
      ]
    ]);
    


val APPEND_EQ_NIL_left = prove_thm("APPEND_EQ_NIL_left",
  (--`!(ys:('a)list)(xs:('a)list) .
     (xs = (APPEND ys xs)) ==> (ys = [])`--),
     
REPEAT STRIP_TAC 
THEN IMP_RES_TAC LENGTH_EQ
THEN REWRITE_ASM_TAC [LENGTH_APPEND]
THEN ASM_CASES_TAC (--`(ys:('a)list) = []`--) THEN ASM_REWRITE_TAC[]
THEN REWRITE_ASM_TAC[GSYM LENGTH_NIL]
THEN UNDISCH_ONE_TAC THEN UNDISCH_ONE_TAC
THEN CONV_TAC ARITH_CONV);

val APPEND_EQ_NIL_right = prove_thm("APPEND_EQ_NIL_right",
  (--`!(ys:('a)list)(xs:('a)list) .
     (xs = (APPEND xs ys)) ==> (ys = [])`--),
     
REPEAT STRIP_TAC 
THEN IMP_RES_TAC LENGTH_EQ
THEN REWRITE_ASM_TAC [LENGTH_APPEND]
THEN ASM_CASES_TAC (--`(ys:('a)list) = []`--) THEN ASM_REWRITE_TAC[]
THEN REWRITE_ASM_TAC[GSYM LENGTH_NIL]
THEN UNDISCH_ONE_TAC THEN UNDISCH_ONE_TAC
THEN CONV_TAC ARITH_CONV);
 
    
   
val APPEND_EQ_thm = prove_thm("APPEND_EQ_thm",
  (--`!(xs:('a)list) (ys:('a)list) (zs:('a)list). 
      ((APPEND xs ys) = (APPEND zs ys)) ==> (xs = zs)`--),

LIST_INDUCT_TAC 
THENL
 [REWRITE_TAC[APPEND] THEN REPEAT STRIP_TAC THEN IMP_RES_TAC APPEND_EQ_NIL_left THEN ASM_REWRITE_TAC[]
 ,
  GEN_TAC THEN GEN_TAC THEN LIST_INDUCT_TAC
  THENL
   [REWRITE_TAC[APPEND_NIL] THEN REPEAT STRIP_TAC 
    THEN POP_ASSUM (fn thm => ASSUME_TAC (SYM thm))
    THEN IMP_RES_TAC APPEND_EQ_NIL_left
   ,
    REWRITE_TAC[APPEND,CONS_11] THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[]
    THEN RES_TAC
   ]
 ]);
     
      

val SUM_1_EQ_LENGTH = prove_thm("SUM_1_EQ_LENGTH",
  (--`!(f:'a->num) (xs:('a)list).
     (!x. (IS_EL x xs) ==> (f x = 1))
     ==>
     ((FOLDR $+ 0 (MAP f xs)) = (LENGTH xs))`--),
     
GEN_TAC THEN LIST_INDUCT_TAC 
THENL[REWRITE_TAC[MAP, LENGTH, FOLDR],ALL_TAC]
THEN REWRITE_TAC[IS_EL, MAP, LENGTH, FOLDR]
THEN REPEAT STRIP_TAC 
THEN SUBGOAL_THEN (--`f (h:'a) = 1`--) (fn thm => ONCE_REWRITE_TAC[theorem "arithmetic" "ADD_SYM"] THEN REWRITE_TAC[thm])
THENL[FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[],ALL_TAC]
THEN REWRITE_TAC[theorem "arithmetic" "ADD1",theorem "arithmetic" "EQ_MONO_ADD_EQ"]
THEN FIRST_ASSUM MATCH_MP_TAC THEN REPEAT STRIP_TAC THEN FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[]);



val SUM_1_OR_0_LESS_LENGTH = prove_thm("SUM_1_OR_0_LESS_LENGTH",

(--`!(f:'a->num) (xs:('a)list).
    (!x. (IS_EL x xs) ==> ((f x = 0) \/ (f x = 1)))
    ==>
    ((FOLDR $+ 0 (MAP f xs)) <= (LENGTH xs))`--),
    
    
GEN_TAC THEN LIST_INDUCT_TAC 
THENL[REWRITE_TAC[IS_EL,MAP, LENGTH, FOLDR] THEN CONV_TAC ARITH_CONV ,ALL_TAC]
THEN REWRITE_TAC[IS_EL,MAP, LENGTH, FOLDR]
THEN REPEAT STRIP_TAC
THEN SUBGOAL_THEN (--`(!x. IS_EL (x:'a) xs ==> (f x = 0) \/ (f x = 1))`--) (fn thm => REWRITE_ASM_TAC[thm])
THENL[GEN_TAC THEN STRIP_TAC THEN FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[], ALL_TAC]
THEN XRULE_ASSUM_TAC (ISPEC (--`h:'a`--)) THEN REWRITE_ASM_TAC[]
THEN UNDISCH_ONE_TAC THEN STRIP_TAC THEN ASM_REWRITE_TAC[theorem "arithmetic" "ADD_CLAUSES"]
THEN ONCE_REWRITE_TAC[theorem "arithmetic" "ADD_SYM"]
THEN REWRITE_TAC[theorem "arithmetic" "ADD1",theorem "arithmetic" "EQ_MONO_ADD_EQ"]
THEN POP_ASSUM (fn thm => ALL_TAC) THEN UNDISCH_ONE_TAC THEN CONV_TAC ARITH_CONV );


val SUM_1_OR_0_AND_ONE_0_LESS_LENGTH = prove_thm("SUM_1_OR_0_AND_ONE_0_LESS_LENGTH",

(--`!(f:'a->num) (xs:('a)list).
    ((!x. (IS_EL x xs) ==> ((f x = 0) \/ (f x = 1))) /\ (?x. (IS_EL x xs) /\ (f x = 0)))
    ==>
    ((FOLDR $+ 0 (MAP f xs)) < (LENGTH xs))`--),
    
    
GEN_TAC THEN LIST_INDUCT_TAC 
THENL[REWRITE_TAC[IS_EL,MAP, LENGTH, FOLDR] THEN CONV_TAC ARITH_CONV ,ALL_TAC]
THEN REWRITE_TAC[IS_EL,MAP, LENGTH, FOLDR]
THEN REPEAT STRIP_TAC
THENL
[SUBGOAL_THEN (--`(!x. IS_EL (x:'a) xs ==> (f x = 0) \/ (f x = 1))`--) (fn thm => REWRITE_ASM_TAC[thm])
THENL[GEN_TAC THEN STRIP_TAC THEN FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[], ALL_TAC]
THEN POP_ASSUM (fn thm => (POP_ASSUM (fn thm => REWRITE_TAC[GSYM thm])) THEN REWRITE_TAC[thm])
THEN REWRITE_TAC[theorem "arithmetic" "ADD_CLAUSES"]
THEN ASM_CASES_TAC (--`?x. IS_EL (x:'a) xs /\ (f x = 0)`--)
THENL
 [POP_ASSUM (fn thm => REWRITE_ASM_TAC[thm]) 
  THEN POP_ASSUM (fn thm => ALL_TAC) 
  THEN UNDISCH_ONE_TAC THEN CONV_TAC ARITH_CONV
 ,
  SUBGOAL_THEN (--`!(x:'a). (IS_EL x xs) ==> (f x = 1)`--) ASSUME_TAC
  THENL[POP_ASSUM (fn thm => ASSUME_TAC (CONV_RULE NOT_EXISTS_CONV thm))
  THEN REPEAT STRIP_TAC THEN XRULE_ASSUM_TAC (ISPEC (--`x:'a`--))
  THEN POP_ASSUM (fn thm => REWRITE_ASM_TAC[thm]) THEN UNDISCH_ONE_TAC
  THEN REWRITE_TAC[IMP_DISJ_THM, DE_MORGAN_THM] THEN ASM_REWRITE_TAC[],ALL_TAC]
  THEN IMP_RES_TAC SUM_1_EQ_LENGTH THEN ASM_REWRITE_TAC[] THEN CONV_TAC ARITH_CONV
 ]
,ALL_TAC]
THEN SUBGOAL_THEN (--`(!x. IS_EL (x:'a) xs ==> (f x = 0) \/ (f x = 1))`--) (fn thm => REWRITE_ASM_TAC[thm])
THENL[GEN_TAC THEN STRIP_TAC THEN FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[], ALL_TAC]
THEN SUBGOAL_THEN (--`?(x:'a). IS_EL x xs /\ (f x = 0)`--) (fn thm => REWRITE_ASM_TAC[thm])
THENL[ EXISTS_TAC (--`x:'a`--) THEN ASM_REWRITE_TAC[], ALL_TAC]
THEN SUBGOAL_THEN (--`(f (h:'a) = 0) \/ (f h = 1)`--) STRIP_ASSUME_TAC
THENL[FIRST_ASSUM MATCH_MP_TAC THEN REWRITE_TAC[],
ASM_REWRITE_TAC[theorem "arithmetic" "ADD_CLAUSES"]
THEN DO_TAC (POP_ASSUM (fn thm => ALL_TAC)) 4 THEN UNDISCH_ONE_TAC  THEN CONV_TAC ARITH_CONV,
ASM_REWRITE_TAC[] THEN ONCE_REWRITE_TAC[theorem "arithmetic" "ADD_SYM"]
THEN ASM_REWRITE_TAC[theorem "arithmetic" "ADD1",theorem "arithmetic" "EQ_MONO_ADD_EQ"]
THEN DO_TAC (POP_ASSUM (fn thm => ALL_TAC)) 4 THEN UNDISCH_ONE_TAC  THEN CONV_TAC ARITH_CONV
]);


val SUM_EQ_LENGTH_AND_0_OR_1_IMP_ALL_1 = prove_thm("SUM_EQ_LENGTH_AND_0_OR_1_IMP_ALL_1",
  (--`!(f:'a->num) (xs:('a)list).
     (((FOLDR $+ 0 (MAP f xs)) = (LENGTH xs))
      /\
      (!x. (IS_EL x xs) ==> ((f x = 0) \/ (f x = 1))))
     ==>
     (!x. (IS_EL x xs) ==> (f x = 1))`--),
     
GEN_TAC THEN LIST_INDUCT_TAC 
THENL[REWRITE_TAC[IS_EL,MAP, LENGTH, FOLDR],ALL_TAC]
THEN REWRITE_TAC[MAP, LENGTH, FOLDR]
THEN REPEAT STRIP_TAC
THEN SUBGOAL_THEN (--`(!x. IS_EL (x:'a) xs ==> (f x = 0) \/ (f x = 1))`--) ASSUME_TAC
THENL[GEN_TAC THEN STRIP_TAC THEN FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[IS_EL],ALL_TAC]
THEN SUBGOAL_THEN (--`(f (h:'a) = 0) \/ (f h = 1)`--) ASSUME_TAC
THENL[XRULE_ASSUM_TAC (ISPEC (--`h:'a`--)) THEN REWRITE_ASM_TAC[IS_EL] THEN ASM_REWRITE_TAC[],ALL_TAC]
THEN UNDISCH_ONE_TAC THEN STRIP_TAC
THENL
 [POP_ASSUM (fn thm => REWRITE_ASM_TAC[thm,theorem "arithmetic" "ADD_CLAUSES"]) 
  THEN SUBGOAL_THEN (--`F`--) ASSUME_TAC
  THENL[ALL_TAC, UNDISCH_TAC (--`F`--) THEN REWRITE_TAC[]]
  THEN IMP_RES_TAC SUM_1_OR_0_LESS_LENGTH  THEN POP_ASSUM (fn thm => ALL_TAC)
  THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[] THEN CONV_TAC ARITH_CONV
 ,ALL_TAC]
THEN UNDISCH_ALL_TAC THEN REWRITE_TAC[IS_EL] THEN REPEAT STRIP_TAC THEN ASM_REWRITE_TAC[]
THEN UNDISCH_TAC (--`IS_EL (x:'a) xs`--) THEN SPEC_TAC ((--`x:'a`--), (--`x:'a`--)) THEN FIRST_ASSUM MATCH_MP_TAC 
THEN ASM_REWRITE_TAC[] THEN POP_ASSUM (fn thm => REWRITE_ASM_TAC[thm])
THEN UNDISCH_TAC (--`1 + FOLDR $+ 0 (MAP f (xs:('a)list)) = SUC (LENGTH xs)`--)
THEN ONCE_REWRITE_TAC[theorem "arithmetic" "ADD_SYM"]
THEN REWRITE_TAC[theorem "arithmetic" "ADD1",theorem "arithmetic" "EQ_MONO_ADD_EQ"]);



val SUM_EQ_PRE_LENGTH_AND_0_OR_1_IMP_ONE_0 = prove_thm("SUM_EQ_PRE_LENGTH_AND_0_OR_1_IMP_ONE_0",
  (--`!(f:'a->num) (xs:('a)list).
    ((NO_DUPLICATES xs) /\ 
     (?y. (IS_EL y xs) /\ (f y = 0) /\  (!x. ((IS_EL x xs) /\ ~(x = y)) ==> (f x = 1))))
     ==>
     ((FOLDR $+ 0 (MAP f xs)) + 1 = ((LENGTH xs)))`--),

GEN_TAC THEN LIST_INDUCT_TAC 
THENL[REWRITE_TAC[IS_EL,MAP, LENGTH, FOLDR],ALL_TAC]
THEN REWRITE_TAC[MAP, LENGTH, FOLDR, IS_EL]
THEN REPEAT STRIP_TAC
THENL
 [UNDISCH_TAC (--`(y:'a) = h`--) THEN STRIP_TAC
  THEN POP_ASSUM (fn thm => REWRITE_TAC[SYM thm] THEN REWRITE_ASM_TAC[SYM thm]) 
  THEN ASM_REWRITE_TAC[theorem "arithmetic" "ADD_CLAUSES"]
  THEN REWRITE_TAC[theorem "arithmetic" "ADD1",theorem "arithmetic" "EQ_MONO_ADD_EQ"]
  THEN MATCH_MP_TAC SUM_1_EQ_LENGTH THEN REPEAT STRIP_TAC THEN FIRST_ASSUM MATCH_MP_TAC 
  THEN ASM_REWRITE_TAC[] THEN IMP_RES_TAC NO_DUPLICATES_THM THEN REWRITE_ASM_TAC[DEL]
  THEN XRULE_ASSUM_TAC (ISPEC (--`y:'a`--)) THEN REWRITE_ASM_TAC[IS_EL]
  THEN STRIP_TAC THEN POP_ASSUM (fn thm => REWRITE_ASM_TAC[thm])
  THEN UNDISCH_ONE_TAC THEN REWRITE_TAC[] THEN FIRST_ASSUM MATCH_ACCEPT_TAC 
 ,
  SUBGOAL_THEN (--`(f (h:'a)) = 1`--) (fn thm => REWRITE_TAC[thm])
  THENL[FIRST_ASSUM MATCH_MP_TAC THEN REWRITE_TAC[]
  THEN IMP_RES_TAC NO_DUPLICATES_THM THEN REWRITE_ASM_TAC[DEL]
  THEN XRULE_ASSUM_TAC (ISPEC (--`h:'a`--)) THEN REWRITE_ASM_TAC[IS_EL]
  THEN STRIP_TAC THEN POP_ASSUM (fn thm => REWRITE_ASM_TAC[thm])
  THEN UNDISCH_ONE_TAC THEN REWRITE_TAC[] THEN FIRST_ASSUM MATCH_ACCEPT_TAC,ALL_TAC]
  THEN REWRITE_TAC[theorem "arithmetic" "ADD1",theorem "arithmetic" "EQ_MONO_ADD_EQ"]
  THEN ONCE_REWRITE_TAC[theorem "arithmetic" "ADD_SYM"] THEN FIRST_ASSUM MATCH_MP_TAC
  THEN IMP_RES_TAC NO_DUPLICATES_CONS THEN ASM_REWRITE_TAC[]
  THEN EXISTS_TAC (--`y:'a`--) THEN ASM_REWRITE_TAC[] THEN REPEAT STRIP_TAC
  THEN FIRST_ASSUM MATCH_MP_TAC THEN ASM_REWRITE_TAC[] 
 ]);
 
val NOT_EL_DEL = prove_thm("NOT_EL_DEL",
  (--`!(xs:('a)list) (x:'a).
    ((~(IS_EL x xs)) ==> (xs = (DEL x xs)))`--),
    
 LIST_INDUCT_TAC  THEN REWRITE_TAC[DEL, theorem "List" "IS_EL"] 
 THEN REWRITE_TAC[DE_MORGAN_THM] THEN REPEAT STRIP_TAC 
 THEN XRULE_ASSUM_TAC GSYM 
 THEN ASM_REWRITE_TAC[] THEN XRULE_ASSUM_TAC GSYM 
 THEN REWRITE_TAC[theorem "List" "CONS_11"] THEN RES_TAC
 THEN ASM_REWRITE_TAC[]);
 
 
val IS_EL_DEL = prove_thm("IS_EL_DEL", 
  (--`!(xs:('a)list) (x:'a) (y:'a). (NO_DUPLICATES xs) ==>
     ((IS_EL x (DEL y xs)) = (~(x = y) /\ (IS_EL x xs)))`--),
  
 LIST_INDUCT_TAC THEN REWRITE_TAC[DEL, theorem "List" "IS_EL"] 
 THEN REPEAT STRIP_TAC THEN IMP_RES_TAC NO_DUPLICATES_CONS THEN IMP_RES_TAC NO_DUPLICATES_THM
 
 THEN REWRITE_ASM_TAC[DEL, theorem "List" "IS_EL"]
 THEN COND_CASES_TAC
 THENL
  [XRULE_ASSUM_TAC (ISPEC (--`y:'a`--)) THEN POP_ASSUM (fn thm => REWRITE_ASM_TAC[SYM thm] THEN REWRITE_TAC[SYM thm])
   THEN EQ_TAC THEN STRIP_TAC 
   THENL
   [ASM_REWRITE_TAC[] THEN STRIP_TAC THEN POP_ASSUM (fn thm => REWRITE_ASM_TAC[SYM thm])
    THEN UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[]
   ,UNDISCH_ONE_TAC THEN ASM_REWRITE_TAC[]
   ]
  ,
   REWRITE_TAC[theorem "List" "IS_EL"] THEN EQ_TAC THEN STRIP_TAC
   THENL
    [POP_ASSUM (fn thm => REWRITE_ASM_TAC[SYM thm] THEN ASM_REWRITE_TAC[thm])
    ,RES_TAC THEN ASM_REWRITE_TAC[]
    ,ASM_REWRITE_TAC[]
    ,RES_TAC THEN ASM_REWRITE_TAC[]
    ]
  ]);
  

val NO_DUPLICATES_NUL = prove_thm("NO_DUPLICATES_NUL",
 (--`NO_DUPLICATES ([]:('a)list)`--),
 REWRITE_TAC[NO_DUPLICATES, theorem "List" "LENGTH", theorem "prim_rec" "NOT_LESS_0"]);


val CONS_NO_DUPLICATES = prove_thm("CONS_NO_DUPLICATES",
   (--`!(xs:('a)list) (h:'a).
      ((NO_DUPLICATES xs) /\ ~(IS_EL h xs)) ==> (NO_DUPLICATES (CONS h xs))`--),
      
REWRITE_TAC[NO_DUPLICATES] THEN REPEAT GEN_TAC THEN STRIP_TAC 
THEN INDUCT_TAC THEN REWRITE_TAC[theorem "List" "EL", theorem "List" "HD", theorem "List" "TL"]
THENL
 [GEN_TAC THEN STRIP_TAC 
  THEN POP_ASSUM (fn thm => STRIP_ASSUME_TAC (REWRITE_RULE[GSYM thm] (ISPEC (--`k:num`--) (theorem "arithmetic" "num_CASES"))))
  THEN ASM_REWRITE_TAC[theorem "List" "EL", theorem "List" "HD", theorem "List" "TL"]
  THEN STRIP_TAC THEN UNDISCH_TAC (--`~(IS_EL (h:'a) xs)`--) THEN ASM_REWRITE_TAC[]
  THEN MATCH_MP_TAC (theorem "List" "ELL_IS_EL") THEN UNDISCH_TAC (--`k < LENGTH (CONS (h:'a) xs)`--)
  THEN ASM_REWRITE_TAC[theorem "List" "LENGTH"] THEN CONV_TAC ARITH_CONV
 ,
  INDUCT_TAC THEN REWRITE_TAC[theorem "List" "EL", theorem "List" "HD", theorem "List" "TL",theorem "List" "LENGTH"]
  THENL
   [STRIP_TAC THEN CONV_TAC (ONCE_DEPTH_CONV SYM_CONV) THEN STRIP_TAC THEN UNDISCH_TAC (--`~(IS_EL (h:'a) xs)`--) THEN ASM_REWRITE_TAC[]
    THEN MATCH_MP_TAC (theorem "List" "ELL_IS_EL") THEN UNDISCH_TAC (--`SUC n < SUC (LENGTH (xs:('a)list))`--)
    THEN CONV_TAC ARITH_CONV
   ,
    STRIP_TAC THEN FIRST_ASSUM MATCH_MP_TAC THEN DO_TAC UNDISCH_ONE_TAC 3
    THEN CONV_TAC ARITH_CONV
   ]
 ]);
 


val NO_DUPLICATES_DEL = prove_thm("NO_DUPLICATES_DEL",
  (--`!(xs:('a)list) (x:'a).
    (NO_DUPLICATES xs) ==> (NO_DUPLICATES (DEL x xs))`--),
    
LIST_INDUCT_TAC THEN REWRITE_TAC[NO_DUPLICATES_NUL,DEL]
THEN REPEAT STRIP_TAC THEN COND_CASES_TAC
THEN IMP_RES_TAC NO_DUPLICATES_CONS THEN ASM_REWRITE_TAC[]
THEN MATCH_MP_TAC CONS_NO_DUPLICATES THEN RES_TAC
THEN ASM_REWRITE_TAC[] THEN IMP_RES_TAC NO_DUPLICATES_THM
THEN UNDISCH_TAC (--`NO_DUPLICATES (xs:('a)list)`--)
THEN STRIP_TAC 
THEN POP_ASSUM (fn thm => ASSUME_TAC (REWRITE_RULE[thm] (ISPEC (--`xs:('a)list`--) IS_EL_DEL)))
THEN ASM_REWRITE_TAC[] THEN POP_ASSUM (fn thm => ALL_TAC)
THEN XRULE_ASSUM_TAC (ISPEC (--`h:'a`--)) THEN UNDISCH_ONE_TAC
THEN REWRITE_TAC[theorem "List" "IS_EL", DEL]);

	
export_theory ();
